#!/bin/csh -f

set fn = $1
if ( ( ! -e $fn ) || ( ! -r $fn ) ) then
	echo Cannot read file $fn
	exit 1
endif

if ( $#argv > 2 ) then
  set hl = ( $argv[2-] )
  @ nh = $#hl
else
  set hl = ( `lchost` `lchost` )
  set nh = 2
endif

if ( $nh > 16 ) then
  echo Too many hosts"," cannot handle $nh hosts
endif

if ( -e /usr/etc/ping ) then
set PING=/usr/etc/ping
else
if  ( -e /etc/ping_ultrix ) then
set PING=/etc/ping_ultrix
else
if ( -e /usr/sbin/ping ) then
# might need to change this for some systems
# set PING='/usr/sbin/ping -c 1'
set PING='/usr/sbin/ping'
else
set PING=ping
endif
endif
endif

if ( -e /usr/etc/arp ) then
set ARP=/usr/etc/arp
else
if ( -e /etc/arp ) then
set ARP=/etc/arp
else
if ( -e /usr/sbin/arp ) then
set ARP=/usr/sbin/arp
else
set ARP=arp
endif
endif
endif

set hostp =  ( 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 )
set hoste =  ( 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 )
set hostfe = ( 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 )
set hostn =  ( 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 )
set hostsn = ( 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 )
set hostun = ( 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 )

if ( $?ROMGEN_BASE_PORT) then
  set base = $ROMGEN_BASE_PORT
else
  set base = `random_int 10000`
  if ( $base < 1024 ) @ base += 1024
endif

if ( $?ROMGEN_FAKE_IP_BASE ) then
    set fakeipbase = $ROMGEN_FAKE_IP_BASE
else
   set fakeipbase = 1
endif

set thishost = `lchost`
set netnum = `pnetnum $thishost`
if ( $?ROMGEN_HOST ) then
	set arphost = $ROMGEN_HOST
else
	set arphost = cheltenham
endif

set i = 1
set imo = 1
foreach host ( $hl )
  set hostp[$i] = $base
  set hostn[$i] = `phostnumber $host`
  set j = 1
  set hcount = 0
  while ( $j < $i )
     if ( $hl[$j] == $host ) @ hcount += 1
     @ j += 1
  end
  if ( $hcount > 0 ) then
      set hostsn[$i] = $netnum.$fakeipbase
      @ fakeipbase += 1
  else
      set hostsn[$i] = $hostn[$i]
  endif
  $PING $host
  set hosteth = `$ARP $host`
  if ( $#hosteth > 3 &&  $hosteth[4] != no ) then
	set hoste[$i] = $hosteth[4]
  else
        if ( $#hosteth <= 3) then
	    set hoste[$i] = '01:02:03:04:05:06'
	else
    	  set hosteth = `rsh $arphost /usr/etc/arp $host`
	  if ( $#hosteth > 3 ) then
	    if ( $hosteth[4] != no ) then
	      set hoste[$i] = $hosteth[4]
	    endif
	  else
	    set hosteth = `rsh $arphost /etc/arp $host`
	    if ( $#hosteth > 3 ) then
	      if ( $hosteth[4] != no ) then
		set hoste[$i] = $hosteth[4]
	      endif
	    else
		set hosteth = `rsh $arphost /usr/sbin/arp $host`
		if ( $#hosteth > 3 ) then
		   if ( $hosteth[4] != no ) then
		     set hoste[$i] = $hosteth[4]
		   endif
		else
		  set hoste[$i] = '01:02:03:04:05:06'
	        endif
            endif
	  endif
	endif
  endif
  set hostfe[$i] = `pfakeether "$hostsn[$i]" $base`
  set hostun[$i] = $host
  if ( $i > 1 && $hostun[$i] == $hostun[$imo] ) then 
	set hostun[$i] = $hostun[$i]$i
  else
	@ imo = $i
  endif
  @ base += 1
  @ i += 1
end

set i = 1
set style = ipeth
foreach host ( $hl )
  set linen = 0
  set file = rom.$host.$i
  if ( -e $file ) rm $file
  set line = `promfile $fn $linen`
  set lline = ( $line )
  echo '# this is for host ' $hostun[$i] >> $file
  while ( $#lline > 0 )
    switch ( $lline[1] )
	case simeth:
	  echo simeth $hostp[$i] >> $file
	  set style = simeth
	  breaksw

	case simsimeth:
	  echo simsimeth $hostp[$i] >> $file
	  set style = simsimeth
	  breaksw

	case dns:
	  set j = 1
	  foreach host ( $hl )
	     if ( $style == ipeth ) then
		     echo dns $hostun[$j] $hostn[$j] >> $file
	     else
		     echo dns $hostun[$j] $hostsn[$j] >> $file
	     endif
	     @ j += 1
          end
	  breaksw

	default:
	  set arp = `echo $line | awk '$1 ~ /arp\// { print $1 } $1 ~ /^arp$/ { print $1 }'`
	  if ( a$arp != a ) then
	    set j = 1
	    switch ( $style )
	    case ipeth :
	      foreach host ( $hl )
	        echo $arp $hostn[$j] $hoste[$j] >> $file
	        @ j += 1
	      end
	      breaksw
	    case simeth :
	       foreach host ( $hl )
	         echo $arp $hostsn[$j] $hostn[$j] $hostp[$j] >> $file
		 echo '#' dns $hostun[$j] $hostfe[$j] >> $file
	         @ j += 1
	       end
	       breaksw
	     case simsimeth :
	       foreach host ( $hl )
		 echo $arp $hostn[$j] $hostfe[$j] >> $file
	         @ j += 1
	       end
	       breaksw 
	     endsw
	  else
	    echo $line >> $file
	  endif
	  breaksw
	endsw
	@ linen += 1
	set line = `promfile $fn $linen`
        set lline = ( $line )
  end
  @ i += 1
end
