/* 
 * $RCSfile: arp_rom.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 *
 *
 * $Log: arp_rom.c,v $
 * Revision 1.2.1.3  1996/11/04 22:52:16  mjk
 * added BOGUS_ARP for linux gateway configuration
 *
 * Revision 1.2.1.2  1996/07/23 20:19:49  mjk
 * allow bogus arp for Linux out-of-kernel stack
 *
 * Revision 1.2.1.1  1994/03/14  23:15:25  umass
 * Uses MAC48BitHosts instead of ETHhosts
 *
 */

/*
 * initialize table from ROM entries
 */

#include "xkernel.h"
#include "arp.h"
#include "arp_i.h"
#include "arp_table.h"
#include "romopt.h"
#include "sim.h"

#ifdef __STDC__

static xkern_return_t loadEntry( XObj, char **, int, int, VOID * );

#else

static xkern_return_t loadEntry();

#endif



static XObjRomOpt	arpOpts[] = {
    { "", 3, loadEntry }
};
 
				/* With -DARP_BOGUSARP the rom file
                                   can specifiy "*" as the IP and have
                                   every ARP request return the same
                                   ethernet address.  All lookups will
                                   consult the table first and return
                                   the default bogus address iff the
                                   table lookup fails.*/
static xkern_return_t
loadEntry( self, str, nFields, line, arg )
    XObj	self;
    char	**str;
    int		nFields, line;
    VOID	*arg;
{
    MAC48bithost genericHost;
    IPhost	 ipHost;
    PSTATE	 *ps = (PSTATE *)self->state;
    int		 nextField;
#ifdef ARP_BOGUSARP
    int		 bogusp = 0;

    if ( !strcmp("*", str[1]) )
      bogusp = 1;
    else
#endif      
      if ( str2ipHost(&ipHost, str[1]) == XK_FAILURE )
	return XK_FAILURE;

    if ( str2mac48bitHost(&genericHost, str[2]) == XK_SUCCESS ) {
	nextField = 3;
    } else {
	/* 
	 * Second field isn't an Ethernet/FDDI address.  See if it's
	 * one of the alternate ways of specifying a hardware address
	 * (there is currently only one alternate)
	 */
	{
	    /* 
	     * Look for SIM address, an IP-host/UDP-port pair
	     */
	    SimAddrBuf	buf;

	    if ( nFields < 4 ) return XK_FAILURE;
	    if ( str2ipHost(&buf.ipHost, str[2]) == XK_FAILURE ) {
		return XK_FAILURE;
	    }
	    if ( sscanf(str[3], "%d", &buf.udpPort) < 1 ) {
		return XK_FAILURE;
	    }
	    if ( xControl(xGetDown(self, 0), SIM_SOCK2ADDR,
			  (char *)&buf, sizeof(buf)) < 0 ) {
		xTraceP0(self, TR_ERRORS, "llp couldn't translate rom entry");
		return XK_FAILURE;
	    }
	    bcopy((char *)&buf.genericHost, (char *)&genericHost, 
	          sizeof(MAC48bithost));
	    nextField = 4;
	}
    }
    
#ifdef ARP_BOGUSARP
    if ( bogusp ) {
      ps->bogusp        = 1;
      ps->bogus_macaddr = genericHost;
      xTraceP1(self, TR_MAJOR_EVENTS, "bogus ETH/FDDI address: %s",
	       mac48bitHostStr(&genericHost));
      return XK_SUCCESS;
    }
#endif
    
    arpSaveBinding( ps->tbl, &ipHost, &genericHost );
    xTraceP1(self, TR_MAJOR_EVENTS, "loaded (%s) from rom file",
	    ipHostStr(&ipHost));
    xTraceP1(self, TR_MAJOR_EVENTS, "corresponding ETH/FDDI address: %s",
	    mac48bitHostStr(&genericHost));

    if ( nFields > nextField ) {
	if ( ! strcmp(str[nextField], "lock") ) {
	    arpLock(ps->tbl, &ipHost);
	} else {
	    return XK_FAILURE;
	}
    }
    return XK_SUCCESS;
}


void
arpPlatformInit( self )
    XObj self;
{
    /*
     * Check the rom file for arp initialization 
     */
    findXObjRomOpts(self, arpOpts, sizeof(arpOpts)/sizeof(XObjRomOpt), 0);
}
