/* 
 * idea_i.h
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993  Arizona Board of Regents
 *
 */

/* #ifndef idea_i_h */
#define idea_i_h

#define MOST_PARTS 5

#define TRANSPORT_PROTOCOL 0
#define KEY_MGR_OUT	   1
#define KEY_MGR_IN	   2
#define KEY_MGR		   KEY_MGR_OUT

typedef long 	CRYPThdr;
typedef long 	CRYPTtype;
typedef XObj  	PassiveId; 

/* private flag names: must stay in sync with public names CRYPT_xxx */
#define IP_PSEUDOHDR_HACK  1  /* either protocol or session */
#define IV_FROM_KEY        2  /* protocol or session */
#define ALL_PARTS          4  /* protocol */
#define CBC                8  /* protocol */
#define DECRYPT           16  /* protocol */
#define USE_IPSEC         32  /* protocol */
#define SEND_IV           64  /* protocol or session */
#define SEND_IV_ONCE     128  /* protocol; copied to session */
#define RCV_IV_ONCE      256  /* protocol; copied to session */
#define CRYPT_IV         512  /* protocol or session */
#define ROTATE_BYTE     1024  /* protocol or session */
#define IV_STATE        2048  /* Save IV after encryption/decryption. PorS */
#define RANDOM_IV       4096  /* protocol or session */
#define IV_FROM_MSG     8192  /* protocol or session */
#define IV_ON_UPMSG    16384  /* protocol or session */
#define FRAGMENTS_DOWN 32768  /* protocol or session */
#define FRAGMENTS_UP   65536  /* protocol or session */
#define NO_PADDING_CHECK   131072  /* protocol */
#define RFC_1829      262144  /* protocol */

#define HEXSTRING     524288  /* protocol */
#define ECB          1048576  /* protocol */
#define CFB          2097152  /* protocol */
#define OFB          4194304  /* protocol */


typedef struct {
    Enable	e;
    XObj	kms;
    XObj	kms2;
    Part	p[2];
    long	method;
    long	flags;    /* &1  Turns on IP pseudoheader length fixup.
			     &2  Turns on Ivec from Key.                */
    u_char      iv_down[8];
    u_char      iv_up[8];
    u_char      iv_random[16];
    u_char      frag_down[8];
    u_char      frag_up[8];
    int         frag_down_ct;
    int         frag_up_ct;
} SState;    /* session state */


typedef struct {
    u_char	deskey[8];
    u_char	descbcvec[8];
} DesCbcData;


typedef XObj ActiveId;

typedef struct {
    Map		actMap;
    Map 	pasMap;
    long	flags;    /* &1  Turns on IP pseudoheader length fixup.
			     &2  Turns on Ivec from Key for all sessions.
			     &4  Turns on using all participants for key lu. 
                             &8  Sets CBC mode.
                             &16 Sets decryption mode. 
                             &32 Sets ipsec mode. 
                          */
    XObj	kms;
    int         interleave; 
} PState;    /* protocol state */

#define	CRYPT_ACTIVE_MAP_SZ	257
#define	CRYPT_PASSIVE_MAP_SZ	13

/* #endif */
