/*
 * $RCSfile: init.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 *
 *
 * $Log: init.c,v $
 * Revision 1.36.2.1  1994/06/01 18:23:20  menze
 * Added definition of rom array for inkernel portion
 *
 * Revision 1.36  1994/02/05  00:30:04  menze
 *   [ 1994/01/14          menze ]
 *   Moved rom initialization stuff to pi/initrom.c
 *
 */

/*
 * Mach3
 */

#include "process.h"
#include "upi.h"
#include "xk_debug.h"
#ifndef XKMACHKERNEL
#include <stdio.h>
#endif XKMACHKERNEL
#include "platform.h"
#include "netmask.h"
#include "prottbl.h"
#include "x_util.h"
#include "rom.h"

/* the event scheduling interval granularity */
int  event_granularity = 100; 

int traceinit = 0;

extern build_pgraph();

#ifndef XKMACHKERNEL

#  ifdef __STDC__

int	main( int, char ** );

#  endif

#endif XKMACHKERNEL

#ifdef MACH
#define SHEPNUM 10 
#endif


/* extern	int 	_init_user(); */

int globalArgc;
char **globalArgv;


#ifdef XKMACHKERNEL

char *rom[ROM_MAX_LINES + 1][ROM_MAX_FIELDS + 1];  

/*
 * Maximum number of args:
 */
#define NMAXARGS	16

/* initialization of testArgs determines default args: */
char *testArgs = "";


static char *testArgv[NMAXARGS]	= {"xkernel"};
static int testArgc = 1;

/* the inkernel version needs a pre-parsed initRom table */
#include "initRom.c"

/*
 * Split "testArgs" into individual arguments and store them in
 * testArgv[]; testArgc is the number of args in testArgv[];
 * testArgv[0] is always "xkernel".
 */
static void
split_args()
{
    extern char* index();
    char *cp = testArgs;

    /* split args and store them in testArgc/testArgv: */

    while (cp && *cp) {
	testArgv[testArgc++] = cp;
	if (testArgc >= NMAXARGS) {
	    Kabort("init.c: too many actual args (increase NMAXARGS)");
	} /* if */
	cp = index(cp, ' ');
	if (cp) {
	    *cp++ = '\0';
	    while (*cp == ' ') {
		++cp;
	    } /* while */
	} /* if */
    } /* while */
    testArgv[testArgc] = 0;
} /* split_args */

#endif XKMACHKERNEL

int
#ifndef XKMACHKERNEL
main(argc,argv)
#else
xkInit(argc, argv)
#endif
int	argc;
char	*argv[];
{
#ifndef XKMACHKERNEL    
    setbuf(stdout, 0);
    globalArgc = argc;
    globalArgv = argv;
#else
    xError("x-kernel initializes");
    split_args();

    globalArgc = testArgc;
    globalArgv = testArgv;
#endif ! XKMACHKERNEL    

    xTraceInit();
    initRom();
#ifdef XKMACHKERNEL
    addRomFlags();
#endif
    map_init();
    msgInit();
    threadInit();
    evInit(event_granularity);
    inputThreadInit();
    xTrace0(init, TR_EVENTS, "Calling initRom");
    netMaskInit();
    xMallocInit();
    prottbl_init();
    upiInit();
    xTrace1(init, TR_EVENTS,
	    "Creating thread to build protocol graph (func %x)",
	    build_pgraph );
    build_pgraph_dev();
    CreateProcess( build_pgraph, STD_PRIO, 0 );
    
    xTrace0(init, TR_EVENTS, "init thread exits");

#ifndef XKMACHKERNEL
    cthread_exit((void *)0);
    return 0;
#else
    thread_terminate( current_thread() );
    thread_halt_self();
    panic( "The zombie xkInit strolls" );
#endif XKMACHKERNEL
}

