#include "xkernel.h"
#include "dheckxleaf.h"
#include "dhkxleaf_i.h"
#include <stdlib.h>
#include "gmp.h"      /* definition of type MP_INT */

int dhkxleafp;
int tracedhkxleafp;

#define EXPO_SIZE 8

void dhkxleaf_init( XObj );
static XObj		dhkxleafOpen( XObj, XObj, XObj, Part * );
static xkern_return_t	dhkxleafOpenDone( XObj, XObj, XObj, XObj );
static xkern_return_t	dhkxleafPop( XObj, XObj, Msg *, void * );
static xkern_return_t	dhkxleafDemux( XObj, XObj, Msg * );
static void		dhkxleafSessnInit(XObj);
static xkern_return_t	dhkxleafClose( XObj );
static int		dhkxleafControl(XObj, int, char *, int);
static void		zap_MP_INT(MP_INT *);
static void		setbignum(MP_INT *, unsigned long *, long);
static void		arrayfrombignum(unsigned long *, MP_INT *, long);
static void		xkFreeZ(void *, size_t);
static void *		xkReallocZ(void *, size_t, size_t);


char * prime_g1 =  "97f64261cab505dd2828e13f1d68b6d3"
         	   "dbd0f313047f40e856da58cb13b8a1bf"
         	   "2b783a4c6d59d5f92afc6cff3d693f78"
         	   "b23d4f3160a9502e3efaf7ab5e1ad5a6"
		   "5e554313828da83b9ff2d941dee95689"
         	   "fadaea0936addf1971fe635b20af4703"
         	   "64603c2de059f54b650ad8fa0cf70121"
         	   "c74799d7587132be9b999bb9b787e8ab\0";


/* These three functions are the interface between the xkernel memory allocation
   functions and the GMP allocation requirements.  I've arranged for the
   free and realloc operations to zero out returned memory, for security. */

extern int gmp_memory_setup;  /* One-time flag for mp_set_memory_functions. */
int gmp_memory_setup = 0;  /* $$ Figure out a way to get this declared once. */

/* As long as xMalloc is a function, this define is ok. */
#define xkgmpMalloc xMalloc

static void xkFreeZ(ptr,size) void *ptr; size_t size;
{ bzero(ptr,size); xFree(ptr); }

static void * xkReallocZ(ptr,oldsize,newsize) void *ptr; size_t oldsize,newsize;
{  void * val;
   val = xkgmpMalloc(newsize);
   if (oldsize >= newsize)
   { printf("xkReallocZ: oldsize %d > newsize %d\n", oldsize, newsize);
     Kabort("xkReallocZ"); }
   bcopy(ptr,val,oldsize);
   xkFreeZ(ptr,oldsize);
   return val; }/* Clears out bignum, then zeros lengths & pointer fields. */

static void zap_MP_INT(mpint) MP_INT* mpint;
{  mpz_clear(mpint); bzero((void *)mpint,sizeof(MP_INT));  }

/* Copy a value from an array to a bignum.  Takes care of possible leading 0s.
   Assumes the bignum contains a value, and clears it before overwriting. */
static void setbignum(pbn,arr,l) MP_INT *pbn; unsigned long *arr; long l;
{ mpz_set_ui(pbn,0);
  for (;l>0;arr++,l--) { mpz_mul_2exp(pbn,pbn,32); mpz_add_ui(pbn,pbn,ntohl(*arr)); } }


/* Copy a value from a bignum into an array.  Arrange for the higher order
   part of the bignum to go into the lower part of the array, so this is
   inverse to setbignum.  Doesn't care if there's bignum left over. */
static void arrayfrombignum(arr,pbn,l) MP_INT *pbn; unsigned long *arr; long l;
{ MP_INT m; mpz_init_set(&m,pbn); 
    for (l--;l>=0;l--) { arr[l] = htonl(mpz_get_ui(&m)); mpz_div_2exp(&m,&m,32); }
  zap_MP_INT(&m); }

static int
dhkxleafControl(self, opcode, buf, len)
    XObj self;
    int opcode;
    char *buf;
    int len;
{
   SState	*ss = (SState *)self->state;
   switch (opcode) {
	case DHECKXLEAF_PUBLIC_VALUE:
		{
       			if (len != DEGW*4)
				return XK_FAILURE;
 			arrayfrombignum((unsigned long *)buf,&ss->gx,DEGW);
 			return XK_SUCCESS;
		}
		break;
	case DHECKXLEAF_CALCULATE_SECRET:
		{
   			MP_INT       gy;
   			MP_INT       gxy;
 			unsigned long value[DEGW]; /*needed for allignment*/
     			if (len != DEGW*4)
				return XK_FAILURE;
			mpz_init(&gxy);
			mpz_init(&gy);

/*			{
			int x;
			unsigned char * y;
			printf("Exponent \n");
			arrayfrombignum(value,&ss->exponent,8);
			y=(unsigned char *) value;
			for (x=0; x <=4*7; x ++) printf("%02x ",y[x]);
			printf("\n Prime:");
			arrayfrombignum(value,&ss->prime,32);
			y=(unsigned char *) value;
			for (x=0; x < 4*32; x ++) printf("%02x ",y[x]);
			printf("\n");
			}
*/		
			
			memcpy((char *)value,buf,DEGW*4);
			setbignum(&gy,value,DEGW); 

			mpz_powm(&gxy,&gy,&ss->exponent,&ss->prime); /* gxy = gy^exponent (mod P) */
 			arrayfrombignum(value,&gxy,DEGW);
			memcpy(buf,(char *)value ,DEGW*4);
			zap_MP_INT(&gxy);			
			zap_MP_INT(&gy);			
			return XK_SUCCESS;
		}
		break;	
	case DHECKXLEAF_VALUE_SIZE:
		{
   			if (len != sizeof(int))
				return XK_FAILURE;
			*((int *)buf)=1024;
			return XK_SUCCESS;
		}
		break;
      	default:
	return XK_FAILURE;
    }
}

static xkern_return_t dhkxleafDemux(self, lls, msg)
     XObj self, lls;
     Msg  *msg;
{
	return XK_SUCCESS;
}

static xkern_return_t dhkxleafPop(self, lls, msg, hdr) 	
	XObj  self,lls;
	Msg * msg;
	void * hdr;
{
	return XK_SUCCESS;
}

static xkern_return_t	dhkxleafOpenDone( self, lls, llp, hlpType  )
     XObj self, lls, llp, hlpType;
{
	return XK_SUCCESS;
}

static XObj dhkxleafOpen(self, hlp, hlpType, part)
     XObj self, hlp, hlpType;
     Part *part;
{
    	SState  * ss;
	XObj s;
	int x;
 	MP_INT       g;
	unsigned long expo[EXPO_SIZE];

   	s = xCreateSessn(dhkxleafSessnInit, hlp, hlp, self, 0 , NULL);
	ss =(SState *) xMalloc(sizeof(SState));
    	bzero((void *)ss, sizeof(SState));
	for(x=0; x < EXPO_SIZE; x++) 
	{
		expo[x] =  rand();
	}
 	mpz_init_set_str(&ss->prime,prime_g1,16);
	mpz_init(&ss->gx);
	mpz_init(&ss->exponent);
	setbignum(&ss->exponent,expo,EXPO_SIZE);
	mpz_init_set_ui(&g,2); 
	
				
	mpz_powm(&ss->gx,&g,&ss->exponent,&ss->prime);  /* gx = g^x (mod P) */

	s->state = ss;
	return s;
}

static xkern_return_t
dhkxleafClose( self )
    XObj	self;
{
   	SState	*ss = (SState *)self->state;
	zap_MP_INT(&ss->gx);			
	zap_MP_INT(&ss->exponent);			
	zap_MP_INT(&ss->prime);			
	xFree(self->state);
	xDestroy(self);
	return XK_SUCCESS;
}



void dhkxleaf_init( self )
     XObj self;
{

      	mp_set_memory_functions((void *(*)(size_t))xkgmpMalloc,xkReallocZ,xkFreeZ);
      	gmp_memory_setup = 1;  	
	self->open = dhkxleafOpen;
  	self->control = dhkxleafControl;
  	self->opendone = dhkxleafOpenDone;
  	self->pop = dhkxleafPop;
  	self->demux = dhkxleafDemux;

}

static void
dhkxleafSessnInit(s)
    XObj s;
{ 
 	s->close = dhkxleafClose;
  	s->pop = dhkxleafPop;
	s->control = dhkxleafControl;
}
