/*     
 * $RCSfile: ip_input.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 *
 *
 * $Log: ip_input.c,v $
 * Revision 1.11.1.6  1996/01/20 06:56:39  ho
 * Support for tunnel mode (recursive call to ip demux, essentially).
 *
 * Revision 1.11.1.1  1994/06/08  17:04:04  menze
 * bzero multicomponent keys before use
 *
 */


#include "xkernel.h"
#include "ip.h"
#include "ip_i.h"
#include "route.h"
#ifdef IPSEC
#include "saidmgr.h"
#include "ipsec-extras.h"
#include "ipsec-combo.h"
extern xkern_return_t ipsecDemux(XObj, XObj, Msg *);
#endif /* IPSEC */

#ifdef __STDC__
static int		onLocalNet( XObj, IPhost * );
static int		validateOpenEnable( XObj );
#ifdef IPSEC
static xmsg_handle_t    ipIpsecCollect (XObj, XObj, Msg *, IPheader *, char *);
#endif /* IPSEC */
#else
static int		onLocalNet();
static int		validateOpenEnable();
#ifdef IPSEC
static xmsg_handle_t    ipIpsecCollect ( );
#endif /* IPSEC */
#endif /* __STDC__ */

/*
 * ipDemux
 */
xkern_return_t
ipDemux(self, transport_s, dg)
    XObj self;
    XObj transport_s;
    Msg *dg;
{
    IPheader	hdr;
    XObj        s;
    ActiveId	actKey;
    PState	*pstate = (PState *) self->state;
    int		dataLen;
    char	options[40];
#ifdef IPSEC
    char        ipnethdr[IPHLEN];
    static long     ipsec_ahid = 0;
    static long     ipsec_espid = 0;

    if (!ipsec_ahid) ipsec_ahid = relProtNumById(protTblGetId("ipsec-ah"),self);
    if (!ipsec_espid) ipsec_espid = relProtNumById(protTblGetId("ipsec-esp"), self);
#endif /* IPSEC */
    xTrace1(ipp, TR_EVENTS,
	    "IP demux called with datagram of len %d", msgLen(dg));
#ifdef IPSEC
    {
      char       *htmp;
      htmp = msgPeek(dg, IPHLEN, NULL);
      bcopy(htmp, ipnethdr, IPHLEN);
    }
#endif /* IPSEC */

    if ( ipGetHdr(dg, &hdr, options) ) {
	xTrace0(ipp, TR_SOFT_ERRORS,
		"IP demux : getHdr problems, dropping message\n"); 
	return XK_SUCCESS;
    }
    xTrace3(ipp, TR_MORE_EVENTS,
	    "ipdemux: seq=%d,frag=%d, len=%d", hdr.ident, hdr.frag,
	    msgLen(dg));
    if (GET_HLEN(&hdr) > 5) {
	xTrace0(ipp, TR_SOFT_ERRORS,
		"IP demux: I don't understand options!  Dropping msg");
	return XK_SUCCESS;
    }
    dataLen = hdr.dlen - IPHLEN;
    if ( dataLen < msgLen(dg) ) {
	xTrace1(ipp, TR_MORE_EVENTS,
		"IP demux : truncating right at byte %d", dataLen);
	msgTruncate(dg, dataLen);
    }
    bzero((char *)&actKey, sizeof(ActiveId));
    actKey.protNum = hdr.prot;
    actKey.remote = hdr.source;
    actKey.local = hdr.dest;
    if ( mapResolve(pstate->activeMap, &actKey, &s) == XK_FAILURE ) {
	FwdId	fwdKey;
	IPhost	mask;

	netMaskFind(&mask, &hdr.dest);
	IP_AND(fwdKey, mask, hdr.dest);
	if ( mapResolve(pstate->fwdMap, &fwdKey, &s) == XK_FAILURE ) {
	    xTrace0(ipp, TR_EVENTS, "no active session found");
	    s = ipCreatePassiveSessn(self, transport_s, &actKey, &fwdKey);
	    if ( s == ERR_XOBJ ) {
		xTrace0(ipp, TR_EVENTS, "...dropping the message");
		return XK_SUCCESS;
	      }
	}
    }
#ifdef IPSEC
    if (actKey.protNum == ipsec_ahid || actKey.protNum == ipsec_espid) {
      return ipIpsecCollect(s, transport_s, dg, &hdr, ipnethdr);
    }
#endif /* IPSEC */
    return xPop(s, transport_s, dg, &hdr);
}

#ifdef IPSEC
static xmsg_handle_t
ipIpsecCollect (s, lls, dg, hdr, ipnethdr)
     XObj s, lls;
     Msg *dg;
     IPheader *hdr;
     char *ipnethdr;
{
    ActiveId	actKey;
    void       *attr = 0 ;
    ipipsecinfo auxinfo;
    PState	*ps = (PState *) s->myprotl->state;
    XObj        secs;
    int        dlen;
    
    /* really should xPop in order to get ref count right */
    s->idle = FALSE;
 redo:
    if (!(COMPLETEPACKET(*(IPheader *)hdr))) {
      xkern_return_t ret;
      /* packet should be reassembled */
      ret = ipReassemble(s, lls, dg, hdr);
      if (ret != XK_MSG_REDO) return ret;
      /* not sure if this is compatible with RFC! */
      /*   we undo any evidence of fragmentation  */
#define FRAG1  6
#define FRAG2  7
#define LEN1   2
#define LEN2   3
      ipnethdr[FRAG1] = ipnethdr[FRAG2] = 0;
      hdr->dlen = dlen = msgLen(dg) + IPHLEN;
      ipnethdr[LEN1] = (dlen>>8) & 0xff;  /* high */
      ipnethdr[LEN2] = dlen &0xff;  /* low */
      hdr->frag = 0;
      hdr->dlen = msgLen(dg);
      hdr->checksum = ~ocsum((u_short *)hdr, sizeof(IPheader) / 2);
    }

	/* checksum validate now */
    bzero((char *)&auxinfo, sizeof(auxinfo));
    bzero((char *)&actKey, sizeof(actKey));
    attr = msgGetAttr(dg, 0);
    auxinfo.iphdr = hdr;
    auxinfo.iphdr_net_form = ipnethdr; /* for AH */
    auxinfo.hdr_size = IPHLEN; /* ? */
    msgSetAttr(dg, 0, &auxinfo, sizeof(auxinfo));
    if (ipsecDemux(xGetDown(s->myprotl, IPSEC_PROTL), s, dg) == XK_FAILURE)
      return XK_FAILURE;
    /* the hdr will have been changed to reflect the "real" hlp */
    if (hdr->prot == relProtNumById(protTblGetId("ipinip"), s->myprotl)) {
      IPheader	inner_hdr;
      char	options[40];

      if ( ipGetHdr(dg, &inner_hdr, options) ) {
	xTrace0(ipp, TR_SOFT_ERRORS,
		"IP ipsec demux : getHdr problems, dropping message"); 
	return XMSG_ERR_HANDLE;
      }
      xTrace3(ipp, TR_MORE_EVENTS,
	      "ipdemux: seq=%d,frag=%d, len=%d", inner_hdr.ident, inner_hdr.frag,
	      msgLen(dg));
      if (GET_HLEN(&inner_hdr) > 5) {
	xTrace0(ipp, TR_SOFT_ERRORS,
		"IP demux: I don't understand options!  Dropping msg");
	return XMSG_ERR_HANDLE;
      }
      if (!IP_EQUAL(inner_hdr.source, hdr->source) ||
	  !IP_EQUAL(inner_hdr.dest, hdr->dest)) {
	xTrace0(ipp, TR_SOFT_ERRORS, "ipIpsecCollect dropping message because of header mismatch on addresses");
	  return XMSG_ERR_HANDLE;
	}
      bcopy((char *)&inner_hdr, (char *)hdr, sizeof(IPheader));
      if (hdr->prot == relProtNumById(protTblGetId("ipsec-ah"), s->myprotl)
	||
	  hdr->prot == relProtNumById(protTblGetId("ipsec-esp"), s->myprotl))
	goto redo;
    }
    actKey.protNum = hdr->prot;
    actKey.remote = hdr->source;
    actKey.local = hdr->dest;
    actKey.secure = msgGetAttr(dg, 0);
    msgSetAttr(dg, 0, attr, sizeof(attr));
    if ( mapResolve(ps->activeMap, &actKey, &secs) == XK_FAILURE ) {
	FwdId	fwdKey;

/*
	netMaskFind(&mask, &hdr.dest);
	IP_AND(fwdKey, mask, hdr.dest);
	if ( mapResolve(ps->fwdMap, &fwdKey, &secs) == XK_FAILURE ) {
*/
	    xTrace0(ipp, TR_EVENTS, "no active session found");
	    secs = ipCreatePassiveSessn(s->myprotl, lls, &actKey, &fwdKey);
	    if ( secs == ERR_XOBJ ) {
		xTrace0(ipp, TR_EVENTS, "...dropping the message");
		return XK_SUCCESS;
	      }
	    ((SState *)(secs->state))->ipsec = actKey.secure;
/*	} */
    }
    if (!(COMPLETEPACKET(*(IPheader *)hdr)))
      {
	xTrace0(ipp, TR_ERRORS, "ipsec in ip: got incomplete encapsulated packet! help!");
	/* return ipReassemble(s, lls, dg, hdr); */
    }
    return ipMsgComplete(secs, lls, dg, hdr);
  }
#endif /* IPSEC */

xkern_return_t
ipForwardPop( s, lls, msg, inHdr )
    XObj	s, lls;
    Msg 	*msg;
    VOID	*inHdr;
{
    IPheader		*h = (IPheader *)inHdr;
    xmsg_handle_t	res;

    xTrace0(ipp, TR_EVENTS, "ip forward pop");
    xAssert(h);
    if ( --h->time == 0 ) {
	xTrace0(ipp, TR_EVENTS, "ttl == 0 -- dropping");
	return XK_SUCCESS;
    }
    /* 
     * We need to go through ipSend because the MTU on the outgoing
     * interface might be less than the packet size (and need
     * fragmentation.) 
     */
    res = ipSend(s, xGetDown(s, 0), msg, h);
    return ( res == XMSG_ERR_HANDLE ) ? XK_FAILURE : XK_SUCCESS;
}


static int
onLocalNet( llo, h )
    XObj	llo;
    IPhost	*h;
{
    int	res;

    res = xControl(llo, VNET_HOSTONLOCALNET, (char *)h, sizeof(IPhost));
    if ( res < 0 ) {
	xTrace0(ipp, TR_ERRORS, "ipFwdBcst couldn't do HOSTONLOCALNET on llo");
	return 0;
    }
    return res > 0;
}


/* 
 * Used for ipFwdBcast sessions, sessions which receive network broadcasts
 * in a subnet  environment.  Depending on the incoming interface, the
 * message may  need to be forwarded on other interfaces and locally
 * accepted, or it may be dropped.  See RFC 922.
 *
 * This is not very efficient and is further complicated by the hiding
 * of interfaces in VNET.
 */
xkern_return_t
ipFwdBcastPop( s, llsIn, msg, inHdr )
    XObj	s, llsIn;
    Msg 	*msg;
    VOID	*inHdr;
{
    IPheader		*h = (IPheader *)inHdr;
    xmsg_handle_t	res;
    route		rt;
    Msg			msgCopy;
    VOID		*ifcId;
    XObj		lls;
    xkern_return_t	xkr;
    PState		*ps = (PState *)xMyProtl(s)->state;

    xTrace0(ipp, TR_EVENTS, "ip forward bcast pop");
    xAssert(h);
    if ( --h->time == 0 ) {
	xTrace0(ipp, TR_EVENTS, "ttl == 0 -- dropping");
	return XK_SUCCESS;
    }
    /* 
     * Did this packet come in on the interface we would use to reach
     * the source?  If not, drop the message
     */
    if ( ! onLocalNet(llsIn, &h->source) ) {
	if ( onLocalNet(xGetDown(xMyProtl(s), 0), &h->source ) ) {
	    xTrace0(ipp, TR_EVENTS,
	          "ipFwdBcast gets packet not on direct connection, dropping");
	    return XK_SUCCESS;
	}
	xkr = rt_get( &ps->rtTbl, &h->source, &rt );
	if ( xkr == XK_FAILURE ) {
	    xTrace0(ipp, TR_SOFT_ERRORS, "ipFwdBcast ... no gateway");
	    return XK_SUCCESS;
	}
	if ( ! onLocalNet(llsIn, &rt.gw) ) {
	    xTrace0(ipp, TR_EVENTS,
		    "ipFwdBcast receives packet from dup gw, dropping");
	    return XK_SUCCESS;
	}
    }
    /* 
     * If this is a session used for local delivery (i.e., if the
     * down[0] is valid), send up for local processing (reassemble first)
     */
    if ( xIsSession(xGetDown(s, 0)) ) {
	msgConstructCopy(&msgCopy, msg);
	ipStdPop(s, llsIn, &msgCopy, h);
	msgDestroy(&msgCopy);
    }
    /* 
     * Send this message back out on all appropriate interfaces except
     * the one on which it was received.
     */
    lls = xGetDown(s, 1);
    xAssert(xIsSession(lls));
    xControl(llsIn, VNET_GETINTERFACEID, (char *)&ifcId, sizeof(VOID *));
    if ( xControl(lls, VNET_DISABLEINTERFACE, (char *)&ifcId, sizeof(VOID *))
		< 0 ) {
	xTrace0(ipp, TR_ERRORS,
		"ipFwdBcastPop could not disable lls interface");
	return XK_SUCCESS;
    }
    res = ipSend(s, lls, msg, h);
    xControl(lls, VNET_ENABLEINTERFACE, (char *)&ifcId, sizeof(VOID *));
    return ( res == XMSG_ERR_HANDLE ) ? XK_FAILURE : XK_SUCCESS;
}


/* 
 * validateOpenEnable -- Checks to see if there is still an openEnable for
 * the session and, if so, calls openDone.
 * This is called right before a message is sent up through
 * a session with no external references.  This has to be done
 * because IP sessions
 * can survive beyond removal of all external references. 
 *
 * Returns 1 if an openenable exists, 0 if it doesn't.
 */
static int
validateOpenEnable( s )
    XObj	s;
{
    SState	*ss = (SState *)s->state;
    Enable	*e;

    e = ipFindEnable(xMyProtl(s), ss->hdr.prot, &ss->hdr.source);
    if ( e == ERR_ENABLE ) {
	xTrace1(ipp, TR_MAJOR_EVENTS, "ipValidateOE -- no OE for hlp %d!",
		ss->hdr.prot);
	return 0;
    }
    xOpenDone(e->hlpRcv, s, xMyProtl(s));
    return 1;
}


xkern_return_t
ipMsgComplete(s, lls, dg, inHdr)
    XObj	s, lls;
    Msg 	*dg;
    VOID	*inHdr;
{
    IPheader *h = (IPheader *)inHdr;
    IPpseudoHdr ph;
#ifdef IPSEC
    static long     ipsec_ahid = 0;
    static long     ipsec_espid = 0;
    static long     ipsec_ipipid = 0;


    if (!ipsec_ahid) ipsec_ahid = relProtNumById(protTblGetId("ipsec-ah"),s->myprotl);
    if (!ipsec_espid) ipsec_espid = relProtNumById(protTblGetId("ipsec-esp"), s->myprotl);
    if (!ipsec_ipipid) ipsec_ipipid = relProtNumById(protTblGetId("ipinip"), s->myprotl);
#endif /* IPSEC */
    
    if ( s->rcnt == 1 && ! validateOpenEnable(s) ) {
	return XK_SUCCESS;
    }
    xAssert(h);
#ifdef IPSEC
    if (h->prot == ipsec_ahid || h->prot == ipsec_espid || h->prot == ipsec_ipipid )
      return XK_MSG_REDO;
#endif /* IPSEC */
    ph.src = h->source;
    ph.dst = h->dest;
    ph.zero = 0;
    ph.prot = h->prot;
    ph.len = htons( msgLen(dg) );
#ifndef DELIVER_ALL
    msgSetAttr(dg, 0, (VOID *)&ph, sizeof(IPpseudoHdr));
#endif
    xTrace1(ipp, TR_EVENTS, "IP pop, length = %d", msgLen(dg));
    xAssert(xIsSession(s));
#ifdef IPSEC
    
#endif
    return xDemux(s, dg);
}



xkern_return_t
ipStdPop( s, lls, dg, hdr )
    XObj	s, lls;
    Msg		*dg;
    VOID	*hdr;
{
    s->idle = FALSE;
    if (COMPLETEPACKET(*(IPheader *)hdr)) {
	return ipMsgComplete(s, lls, dg, hdr);
    } else {
	return ipReassemble(s, lls, dg, hdr);
    }
}


