/*     
 * $RCSfile: ipsec_simple.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1994,1993,1991,1990  Arizona Board of Regents
 *
 *
 * $Log: ipsec_simple.c,v $
 * Revision 1.1.1.6  1994/11/16 02:30:28  ho
 * Puts ip protocol number in ipsec hdr; ipsec id is used w/ip.
 *
 * Revision 1.1.1.5  1994/10/20  02:08:01  ho
 * Support for ip broadcast.
 *
 * Revision 1.1.1.1  1994/08/03  00:15:23  menze
 * Fixed prototypes problems
 *
 */

/*
 *  Maps to/from SAID's, for purposes of providing IP security.
 *  Outgoing messages get the remote site's SAID attached and
 *  get encrypted in the key negotiated for that SAID
 *
 *  Incoming messages get decrypted according to the key associated
 *  with the SAID
 *
 *  We assume one active said per src/dst pair.  There can be
 *  multiple upward protocols.  There would be difficulties if
 *  there were algorithm negotiation; we would need to associate
 *  the said,alg pair with an appropriate hlp at demux time.
 *  The algorithm negotiator would have to know which algorithms
 *  are available and how to indicate which one goes with the said.
 *
 */

#include "xkernel.h"
#include "ip.h"
#include "ipsec.h"
#include "ipsec_simple_i.h"

int	traceipsecp;


#ifdef __STDC__
static void		ipsec_getproc_sessn(XObj);
static void		ipsec_getproc_protl(XObj);
static xkern_return_t	ipsecCloseProtl( XObj );
static xkern_return_t	ipsecCloseSessn( XObj );
static int		ipsecControlProtl( XObj, int, char *, int ) ;
/*static int		ipsecControlSessn( XObj, int, char *, int ) ; */
static XObj		ipsecCreateSessn( XObj, XObj, XObj, HostPair *, ActiveId *, said_t, said_t, LOCALHOSTTYPE *);
static xkern_return_t	ipsecDemux( XObj, XObj, Msg * );
static long		ipsecGetSaid( void *, char *, long, void * );
static void		ipsecPutSaid( void *, char *, long, void * );
static XObj		ipsecOpen( XObj, XObj, XObj, Part * );
static xkern_return_t	ipsecOpenDisable( XObj, XObj, XObj, Part * );
static xkern_return_t	ipsecOpenEnable( XObj, XObj, XObj, Part * );
static xkern_return_t	ipsecOpenDisableAll( XObj, XObj );
static xkern_return_t	ipsecOpenDone( XObj, XObj, XObj, XObj);
static xkern_return_t	ipsecPop( XObj, XObj, Msg *, void * );
static xmsg_handle_t	ipsecPush( XObj, Msg * );
static int		xCopyParts(Part **, char **, Part *, XObj);

#else
static void		ipsec_getproc_sessn();
static void		ipsec_getproc_protl();
static xkern_return_t	ipsecCloseProtl();
static xkern_return_t	ipsecCloseSessn();
static int		ipsecControlProtl() ;
/* static int		ipsecControlSessn() ; */
static XObj		ipsecCreateSessn();
static xkern_return_t	ipsecDemux();
static long		ipsecGetSaid();
static void		ipsecPutSaid();
static XObj		ipsecOpen();
static xkern_return_t	ipsecOpenDisable();
static xkern_return_t	ipsecOpenEnable();
static xkern_return_t	ipsecOpenDisableAll();
static xkern_return_t	ipsecOpenDone();
static xkern_return_t	ipsecPop();
static xmsg_handle_t	ipsecPush();
static int		xCopyParts();

#endif /* __STDC__ */

LOCALHOSTTYPE Any_Host = { -1, -1, -1, -1 };

#define myxGetDown(prot, n) ( ((n <= STD_DOWN ) ? prot->down[n] :  xGetDown(prot, n)))

#define xGetProtocolInstanceName(obj, ptr) \
  ( ptr = obj->instName )

#define initMapKey(k) bzero((char *)&(k), sizeof(k))

#define LOCAL_PART_MAX 4*sizeof(LOCALHOSTTYPE)

/*
 * ipsec_init
 */
void
ipsec_init(self)
    XObj self;
{
    PSTATE	*pstate;
    XObj	lltp;
    
    xTrace0(ipsecp, TR_GROSS_EVENTS, "IPSEC init");
    xAssert(xIsProtocol(self));
    
    ipsec_getproc_protl(self);
    pstate = X_NEW(PSTATE);
    bzero((char *)pstate, sizeof(PSTATE));
    self->state = (VOID *)pstate;

    /* 
      active map1 is used for open.
      lookup based on hlp/src/dst, gets lls/said if open
      else add <hlp/src/dst, session>

      active map2 is used for demux
      lookup based on said/src/dst, gets session(hlp) if open
      */
    pstate->activemap = mapCreate(ACTIVE_MAP_SIZE, sizeof(ActiveId));

    /* 
      passive map is used for openenable.
      adds <src/dst, hlp> if not already there

      passive map is used for demux
      lookup based on src/dst, gets hlp if openenabled
         then adds <hlp/src/dst, session> to active map1
              adds <said/src/dst, session> to active map2
      */

    pstate->passivemap = mapCreate(PASSIVE_MAP_SIZE, sizeof(PassiveId));
    if ( ! xIsProtocol(lltp = myxGetDown(self, TRANSPORT_PROTOCOL)) ) {
	Kabort("IPSEC could not get lower transport protocol");
    }
#ifdef OT
    partInit(&part, 1);
    partPush(part, ANY_HOST, 0);
    if ( xOpenEnable(self, self, lltp, &part) == XK_FAILURE ) {
	xTrace0(ipsecp, TR_ALWAYS,
		"ipsec_init: can't openenable transport protocol");
	xFree((char *) pstate);
	return;
    }
#endif
    xTrace0(ipsecp, TR_GROSS_EVENTS, "IPSEC init done");
}

static int
xCopyParts(q, qbuf, p, lls)
     Part **q;
     char **qbuf;
     Part *p;
     XObj lls;
{
  char *ebuf;
  int	len = 500;

  ebuf = xMalloc(500);
  if (lls != ERR_XOBJ) {
    if (xControl(lls, GETPARTICIPANTS, ebuf, len)<0)
      return 0;
  }
  else
    partExternalize(p, ebuf, &len);
  *q = (Part *)xMalloc(partExtLen(ebuf)*sizeof(Part));
  partInternalize(*q, ebuf);
  *qbuf = ebuf;
  return partLen(((Part *)(*q)));
}

/*
 * ipsecOpen
 */
static XObj
ipsecOpen( self, hlpRcv, hlpType, p )
    XObj self, hlpRcv, hlpType;
    Part *p;
{
    ActiveId 	key;
    PSTATE	*pstate = (PSTATE *)(self->state);
    char	*hlp_iname;
    LOCALHOSTTYPE *rh, *lh;

    xTrace0(ipsecp, TR_MAJOR_EVENTS, "IPSEC open");

    pstate = (PSTATE *)self->state;
    initMapKey(key);
    xGetProtocolInstanceName(hlpRcv, hlp_iname);
    /* the instance name should match the global SAID TABLE */
    key.said = atoi(hlp_iname + 1);
    key.prot = relProtNum(hlpType, self);
    if (key.said <= 0) {
      xTrace1(ipsecp, TR_SOFT_ERRORS, "IPSEC Could not get instance number of upper protocol --- %s is invalid", hlp_iname);
      return ERR_XOBJ;
    }

    key.lls = xOpen(self, self, myxGetDown(self, TRANSPORT_PROTOCOL), p);
    if (key.lls != ERR_XOBJ) {
      XObj s;

      xTrace3(ipsecp, TR_DETAILED, "ipsec open said %x lls %x prot %d",
	      key.said, key.lls, key.prot);

      if (mapResolve(pstate->activemap, &key, &s) == XK_SUCCESS)
	{
	  /* already had this, just get another reference */
	  ((SSTATE *)(s->state))->refcnt++;
	  xTrace2(ipsecp, TR_MAJOR_EVENTS, "IPSEC open: reusing session from host %x to host %x",
		  ((SSTATE *)(s->state))->hosts.localhost,
		  ((SSTATE *)(s->state))->hosts.remotehost);

	  return s;
	}
      else {
	Part	*qart;
	char	*b;
	HostPair hosts;

	
	xCopyParts(&qart, &b, 0, key.lls);
	rh = (LOCALHOSTTYPE *)partPop(qart[REMOTE]);
	lh = (LOCALHOSTTYPE *)partPop(qart[LOCAL]);
	if (lh) hosts.localhost = (lh!=ANY_HOST)? *lh:Any_Host;
	if (rh) hosts.remotehost = (rh!=ANY_HOST)? *rh:Any_Host;
	xTrace4(ipsecp, TR_MAJOR_EVENTS, "IPSEC open: new session from host %x to host %x said %x lls %x",
	    hosts.localhost, hosts.remotehost, key.said, key.lls);

	s = ipsecCreateSessn(self, hlpRcv, hlpType, &hosts, &key, key.said, key.said, lh);
	return s;
      }
    }
    return ERR_XOBJ;
  }

static xkern_return_t
ipsecOpenDone( self, lls, llp, hlpType)
     XObj self, lls, llp, hlpType;
{
  /* PSTATE *ps; */

  xTrace0(ipsecp, TR_FULL_TRACE, "ipsec: opendone called, doing nothing yet");
  return XK_SUCCESS;
}

static xkern_return_t
ipsecOpenEnable(s, h, htype, p)
     XObj s, h, htype;
     Part *p;
{
  PassiveId key;
  Part q[1];
  Enable *en;
  char *hlp_iname;
  LOCALHOSTTYPE *rh;
  short pid = relProtNum(htype, s);

  xTrace0(ipsecp, TR_FULL_TRACE, "ipsec: openenable called");
  
  bzero((char *)&key, sizeof(key));

  q[REMOTE] = p[REMOTE];

  /* the only participant will be the remote host */
  rh = (LOCALHOSTTYPE *)partPop(q[REMOTE]);
  if (rh) {
      key.h = (rh==ANY_HOST) ? Any_Host : *rh;
    }
  else {
    xTrace0(ipsecp, TR_ERRORS, "ipsec: no participants in openenable");
    return XK_FAILURE;
  }
  key.prot = pid;
  xGetProtocolInstanceName(h, hlp_iname);
    /* the instance name should match the global SAID TABLE */
  key.sd = atoi(hlp_iname+1);
  if (key.sd <=0 ) {
    xTrace1(ipsecp, TR_ERRORS, "ipsec: cannot get upper protocol subinstance type from %x", hlp_iname);
    if (hlp_iname)
      xTrace1(ipsecp, TR_ERRORS, "ipsec:  upper protocol instance name %s", hlp_iname);
  }
  if ((mapResolve(((PSTATE *)(s->state))->passivemap, &key, &en)==XK_FAILURE)
      ||
      (en->hlpRcv != h))
    {
      en = X_NEW(Enable);
      bzero((char *)en, sizeof(Enable));
      en->hlpRcv = h;
      en->hlpType = htype;
      en->binding = mapBind(((PSTATE *)(s->state))->passivemap, &key, en);
    }
  en->rcnt++;
  return(xOpenEnable(s, s, myxGetDown(s, TRANSPORT_PROTOCOL), p));
}

static XObj
ipsecCreateSessn( XObj self, XObj hlpRcv, XObj hlpType, HostPair *hosts,
		  ActiveId *key, said_t said, said_t remsaid,
		  LOCALHOSTTYPE *rcvhost)
{
    XObj	s;
    SSTATE	*sstate;
    PSTATE	*pstate;

    pstate = (PSTATE *)self->state;
    s = xCreateSessn(ipsec_getproc_sessn,
		     hlpRcv, hlpType, self, 1, &key->lls);
    sstate = (SSTATE *)xMalloc(sizeof(SSTATE));
    bzero((char *)sstate, sizeof(SSTATE));
    s->state = (char *)sstate;
    sstate->local_said = said;
    sstate->remote_said = remsaid;
    sstate->ihdr.said = remsaid;
    sstate->ihdr.prot = key->prot;
    sstate->hosts = *hosts;
    sstate->myhost = *rcvhost;
    sstate->hlp = hlpRcv;
    sstate->refcnt = 1;
    sstate->flags = 0;
    /* key is clean from caller */
    s->binding = mapBind(pstate->activemap, key, s);
/*    xDuplicate(key-lls); */
    return s;
  }

static xkern_return_t
ipsecPop( s, lls, m, hdr)
  Msg *m;
  XObj s, lls;
  VOID *hdr;
{
  /* said_t said= *(said_t *)hdr; */

  if (((SSTATE *)(s->state))->flags & IP_PS_HACK_FLAG
	|| ((PSTATE *)(xMyProtl(s)->state))->flags & IP_PS_HACK_FLAG)
    {  
      IPpseudoHdr *ipphdr = msgGetAttr( m, 0);
      if (ipphdr)
	{  
	  long len;

	  len = msgLen(m);
	  ipphdr->len = htons(ntohs(ipphdr->len) - IPSEC_HDR_NETSIZE);
	  xTrace4( ipsecp, TR_FULL_TRACE,
		   "IPSEC Pop:  Adjusting IP pseudoheader length; ipphdr %x  phlen after %d  len %d  dlen %d",
		   ipphdr, ntohs(ipphdr->len), len, len-IPSEC_HDR_NETSIZE); 
/*	  ipphdr->prot = ((SSTATE *)(s->state))->ihdr.prot; */
	 }
      else { xTrace0( ipsecp, TR_FULL_TRACE,
		     "IPSEC Pop:  Skipping length adjustment for IP pseudoheader - msg attr is 0."); };
    }
  
  return xDemux(s, m);
}

static long ipsecGetSaid(dst, src, len, arg)
     void *dst;
     char *src;
     long len;
     void *arg;
{
  ipsecHdr *dest = (ipsecHdr *)dst;
  short ntord;

  xAssert(len>=IPSEC_HDR_NETSIZE);
/*  *(said_t *)dst = ntohl((long)src); */
  dest->said = *src;
  bcopy(src+1, (char *)&ntord, PROT_NETSIZE);
  dest->prot = ntohs(ntord);
  return IPSEC_HDR_NETSIZE;
}

static void ipsecPutSaid(src, dst, len, arg)
     void *src;
     char *dst;
     long len;
     void *arg;
{
  ipsecHdr *sorc = (ipsecHdr *)src;
  short htord;

  xAssert(len==IPSEC_HDR_NETSIZE);
/*  *(long *)dst = htonl((long)src); */
  *dst = sorc->said;
  htord = htons(sorc->prot);
  bcopy((char *)&htord, dst+1, PROT_NETSIZE);
}

/*
 * ipsecDemux
 */
static xkern_return_t
ipsecDemux(self, transport_s, dg)
    XObj self;
    XObj transport_s;
    Msg *dg;
{
    XObj        s;
    ActiveId	key;
    PassiveId	pkey;
    PSTATE	*pstate = (PSTATE *) self->state;
    HostPair	hosts;
    LOCALHOSTTYPE rcvhost;
    ipsecHdr	ihdr;

    xTrace2(ipsecp, TR_EVENTS,
	    "IPSEC demux called with datagram of len %d, lls %x", msgLen(dg),
	    transport_s);

    initMapKey(key);
    initMapKey(pkey);
    msgPop(dg, ipsecGetSaid, &ihdr, sizeof(ipsecHdr), 0);
    key.lls = transport_s;
    key.said = ihdr.said;
    key.prot = ihdr.prot;
/*
    xControl(transport_s, GETMYHOST, (char *)&actKey.hosts.localhost,
	     sizeof(LOCALHOSTTYPE));
*/
    if ( mapResolve(pstate->activemap, &key, &s) == XK_FAILURE )
    {
      Enable *e;
      LOCALHOSTTYPE rh;

      xControl(transport_s, GETPEERHOST, (char *)&pkey.h,
	       sizeof(LOCALHOSTTYPE));
      pkey.prot = ihdr.prot;
      pkey.sd = key.said;
      if ( mapResolve(pstate->passivemap, &pkey, &e ) == XK_FAILURE )
	{
	  rh = pkey.h;
	  pkey.h = Any_Host;
	  if ( mapResolve(pstate->passivemap, &pkey, &e) == XK_FAILURE) {
	    xTrace2(ipsecp, TR_EVENTS, "ipsec no passive session found for %d %x",
		    pkey.sd, pkey.h);
	    return XK_FAILURE;
	  }
	  pkey.h = rh;
	}
      xTrace2(ipsecp, TR_DETAILED, "ipsec demux new session for %x %x",
	      key.said, key.lls);
      hosts.remotehost = pkey.h;
      {
	char *ebuf = xMalloc(500);
	int len = 500;
	Part p[2];
	LOCALHOSTTYPE *pl, *pr;
	if (
	    xControl(transport_s, GETPARTICIPANTS, ebuf,
		     len) < 0) {
	  xFree(ebuf);
	  return XK_FAILURE;
	}
	else {
	  partInternalize(p, ebuf);
	  pl = partPop(p[LOCAL]);
	  pr = partPop(p[REMOTE]);
	  hosts.localhost = *pl;
	  hosts.remotehost = *pr;
	  xFree(ebuf);
	}
      }
      xControl(transport_s, GETMYHOST, (char *)&rcvhost,
	       sizeof(LOCALHOSTTYPE));
      s = ipsecCreateSessn(self, e->hlpRcv, e->hlpType, &hosts,
			   &key, pkey.sd, pkey.sd, &rcvhost);
      xDuplicate(transport_s);
      xOpenDone(e->hlpRcv, s, self);
    }
    else
      xTrace2(ipsecp, TR_DETAILED, "ipsec demux using existing session for %x %x",
	      key.said, key.lls);

    return xPop(s, transport_s, dg, (void *)0);
}

/*
 * ipsecPush
 */
static xmsg_handle_t
ipsecPush(s, msg)
     XObj s;
     Msg *msg;
{
    SSTATE	*sstate;
    
    xAssert(xIsSession(s));
    sstate = (SSTATE *) s->state;
    
    xTrace1(ipsecp, TR_EVENTS, "ipsec push said %x", sstate->remote_said);
    msgPush(msg, ipsecPutSaid, &sstate->ihdr, IPSEC_HDR_NETSIZE, 0);
    return xPush(xGetDown(s,TRANSPORT_PROTOCOL), msg);
  }


/*
 * ipsecControlProtl
 */
static int
ipsecControlProtl(self, opcode, buf, len)
    XObj self;
    int opcode;
    char *buf;
    int len;
{
    xAssert(xIsProtocol(self)||xIsSession(self));
    
    switch (opcode) {
	
      case GETMAXPACKET:
      case GETOPTPACKET:
	{ 
	    int max;

	    if (len < sizeof(max)) return 0; 
	    if (xControl(myxGetDown(self, 0), opcode, (char *) &max, sizeof(max)) <= 0)
	      {
		return 0; 
	      }
	    max -= sizeof(said_t);
	    *(int *)buf = max;
	    return(sizeof(max));
	}

      case GETMYHOST :
      case IPSEC_GETMYHOST:
	{
	  SSTATE *sstate;
	  
	  xAssert(xIsSession(self)||xIsProtocol(self));
	  if (xIsSession(self)) {
	    sstate = (SSTATE *)self->state;
	    checkLen(len, LOCALHOSTLEN);
	    *(LOCALHOSTTYPE *)buf = sstate->myhost;
	    return LOCALHOSTLEN;
	  }
	  else return -1;
	}
	
      case GETPEERHOST :
      case IPSEC_GETPEERHOST :
	{
	  SSTATE *sstate;

	  xAssert(xIsSession(self));
	  sstate = (SSTATE *)self->state;
	  checkLen(len, LOCALHOSTLEN);
	  *(LOCALHOSTTYPE *)buf = sstate->hosts.remotehost;
	  return LOCALHOSTLEN;
	}
	

      case GETMYHOSTCOUNT:
      case GETPEERHOSTCOUNT:
	checkLen(len, sizeof(int));
	*(int *)buf = 1;
	return sizeof(int);

      case IP_GETPSEUDOHDR:
	{
	  int val;
	  SSTATE *ss = (SSTATE *)self->state;
	  
	  xAssert(xIsSession(self));
	  val = xControl(myxGetDown(self,0), opcode, buf, len);
/*	  ((IPpseudoHdr *)(buf))->prot = ss->ihdr.prot; */
	  return val;
      }
	break;

      case IP_PSEUDOHDR:
	/* Turn on flag saying to fixup length on IPpseudoheader. */
	{   ((PSTATE *)(self->state))->flags |= 1;
	    xTrace0(ipsecp, TR_EVENTS,
	      "ControlProtl: Turning on IPpseudoheader length fixup.");  }
/* **** FALL THROUGH TO DEFAULT CASE, so lower protocols see it too.  **** */
/* **** Code for this control-op should be immediately above default: **** */

      default:
        if (buf && len) bzero(buf, len);   /* prevent info going down */
	return xControl(myxGetDown(self,0), opcode, buf, len);
    }
}

static xkern_return_t
ipsecCloseSessn(s)
     XObj s;
{
  xTrace0(ipsecp, TR_FULL_TRACE, "ipsecCloseSessn called");
  xFree((char *)(s->state));
  mapRemoveBinding(((PSTATE *)(s->myprotl->state))->activemap, s->binding);
  xClose(xGetDown(s, TRANSPORT_PROTOCOL));
  return XK_FAILURE;
}

static xkern_return_t
ipsecCloseProtl(s)
     XObj s;
{
  xTrace0(ipsecp, TR_FULL_TRACE, "ipsecCloseProtl called");
  xTrace0(ipsecp, TR_ERRORS, "ipsecCloseProtl not implemented");
  return XK_FAILURE;
}

static xkern_return_t
  ipsecOpenDisable( s, h1, h2, p)
  XObj s, h1, h2;
  Part *p;
{
  xTrace0(ipsecp, TR_FULL_TRACE, "ipsecOpenDisable called");
  xTrace0(ipsecp, TR_ERRORS, "ipsecOpenDisable not implemented");
  return XK_FAILURE;
}

static xkern_return_t
  ipsecOpenDisableAll( self, h)
  XObj self,h;
{
  xTrace0(ipsecp, TR_FULL_TRACE, "ipsecOpenDisableAll called");
  xTrace0(ipsecp, TR_ERRORS, "ipsecOpenDisableAll not implemented");
  return XK_FAILURE;
}

static void
ipsec_getproc_protl(s)
    XObj s;
{
  xAssert(xIsProtocol(s));
  s->close = ipsecCloseProtl;
  s->control = ipsecControlProtl;
  s->open = ipsecOpen;
  s->openenable = ipsecOpenEnable;
  s->opendisable = ipsecOpenDisable;
  s->opendone = ipsecOpenDone;
  s->demux = ipsecDemux;
  s->opendisableall = ipsecOpenDisableAll;
}

static void
ipsec_getproc_sessn(s)
    XObj s;
{
  xAssert(xIsSession(s));
  s->push = ipsecPush;
  s->pop = ipsecPop;
  s->control = ipsecControlProtl;
  s->close = ipsecCloseSessn;
}
