#include "xkernel.h"
#include <krb5/krb5.h>
#include <isode/psap.h>
#include <krb5/kdb.h>
#include <krb5/asn1.h>
#include <com_err.h>

#include "prot/kerb_services.h"
#include "kerb_services_i.h"
#include "prot/join.h"
#include "prot/kerb_serv_bottom.h"
#include "prot/sendauth.h"
#include "prot/hash.h"
#include "prot/crypt.h"
#include "prot/recvauth.h"

#ifdef __STDC__

void             kerb_services_init( XObj );
XObj             kerb_servOpen( XObj, XObj, XObj, Part * );
xkern_return_t   kerb_servDemux( XObj, XObj, Msg * );
static xkern_return_t	kerb_servOpenEnable( XObj, XObj, XObj, Part * );
static xkern_return_t	kerb_servOpenDone( XObj, XObj, XObj, XObj);
int              kerb_servProtControl( XObj, int, char *, int );
int              kerb_servSessnControl( XObj, int, char *, int );
void             sessn_init( XObj );
xmsg_handle_t    kerb_servPush( XObj, Msg * );
xkern_return_t   kerb_servPop( XObj, XObj, Msg *, void * ); 
static int       partToId(Part *, char *, int);
static int       part2ToId(Part *, char *, int);
static void      mycopypush(void *, char *, long , void *);


#else


void             kerb_services_init();
XObj             kerb_servOpen();
xkern_return_t   kerb_servDemux();
static xkern_return_t	kerb_servOpenEnable();
static xkern_return_t	kerb_servOpenDone();
int              kerb_servProtControl();
int              kerb_servSessnControl();
void             sessn_init();
xmsg_handle_t    kerb_servPush();
xkern_return_t   kerb_servPop(); 
static int       partToId();
static int       part2ToId();
static void      mycopypush();

#endif



int tracekerb_servicesp = 1;
#define KERBSERVP kerb_servicesp
#define MAP_SZ 32


void kerb_services_init(self)
     XObj self;
{
  self->state = (void *)xMalloc(sizeof(PSTATE));
  if (!self->state) {
    xTrace0(KERBSERVP, TR_ERRORS,
	    "could not malloc pstate for kerb_services");
    Kabort("malloc failed for kerb_services state");
  }
  bzero((char*)self->state, sizeof(PSTATE));

  ((PSTATE*)self->state)->activemap = mapCreate(MAP_SZ, sizeof(XObj));
  ((PSTATE*)self->state)->default_enhancement = None;


  self->open = kerb_servOpen;
  self->demux = kerb_servDemux;
  self->control = kerb_servProtControl;
  self->openenable = kerb_servOpenEnable;
  self->opendone = kerb_servOpenDone;

}



XObj kerb_servOpen(self, hlpType, hlpRcv, part)
     XObj self, hlpType, hlpRcv;
     Part *part;
{
  PSTATE        *ps = (PSTATE *)self->state;
  SSTATE        *sstate;
  SendauthArgs  *args;
  char           partBuf[256];
  XObj           s;
  int            i;

  xTrace0(KERBSERVP, TR_EVENTS, "kerb_servOpen -- entered");

  if (partLen(part) < 1) {
    xTrace0(KERBSERVP, TR_ERRORS, "bad participant list");
    return ERR_XOBJ;
  }

  /* need to pop off the sendauth args;  put them back
     on the part list only for sendauth protocol */
  args = partPop(*part);

  s = xCreateSessn(sessn_init, hlpRcv, hlpType, self, 0,0);
  sstate = (SSTATE *)xMalloc(sizeof(SSTATE));
  if (!sstate) {
    xTrace0(KERBSERVP, TR_ERRORS, "malloc failed for sstate");
    return ERR_XOBJ;
  }
  s->state = (void *)sstate;
  bzero((char*)sstate, sizeof(SSTATE));
  sstate->enhancement = Private;

  sstate->lower[0] = xOpen(self, self, xGetDown(self, 0), part);
  if (sstate->lower[0] == ERR_XOBJ) {
    xTrace0(KERBSERVP, TR_ERRORS,
	    "Could not open kerb_serv_bottom control session");
    return ERR_XOBJ;
  }

  if (xControl(sstate->lower[0], GETPARTICIPANTS, partBuf, 256) <= 0) {
    xTrace0(KERBSERVP, TR_ERRORS, "could not get participants");
    return ERR_XOBJ;
  }

  mapBind(ps->activemap, &(sstate->lower[0]), s);
  if (xControl(sstate->lower[0], JOINSETCONTROL, (char *)&s, sizeof(XObj)) <= 0) {
    xTrace0(KERBSERVP, TR_ERRORS, "Could not set control session for sendauth");    return ERR_XOBJ;
  }

  for (i = 1; i < 4; i++) {
    partInternalize(part, partBuf);
    sstate->lower[i] = xOpen(self, self, xGetDown(self,i), part);
    if (sstate->lower[i] == ERR_XOBJ) {
      xTrace1(KERBSERVP, TR_ERRORS, "could not open lls %d", i);
      return ERR_XOBJ;
    }
    mapBind(ps->activemap, &sstate->lower[i], s);
  }
  partInternalize(part, partBuf);
  partPush(part[0], args, sizeof(SendauthArgs));
  sstate->lower[4] = xOpen(self, self, xGetDown(self,4), part);
  if (sstate->lower[4] == ERR_XOBJ) {
    xTrace0(KERBSERVP, TR_ERRORS, "could not open lls 4");
    return ERR_XOBJ;
  }
  mapBind(ps->activemap, &sstate->lower[4], s);

  /* since lower[4] did not return ERR_XOBJ, we know that the
     authentication was successful */
  /* we should tell kerb_serv_bottom that we are no longer in
     authentication mode */
  xControl(sstate->lower[0], END_AUTHENTICATION, 0,0);
  
  
  /* pass the key to crypt */
  /* ADD CODE HERE */
  xControl(sstate->lower[3], CRYPT_OUT_SET,
	   (char *)args->credsp->keyblock.contents, args->credsp->keyblock.length); 
  xControl(sstate->lower[3], CRYPT_IN_SET,
	   (char *)args->credsp->keyblock.contents, args->credsp->keyblock.length); 
  xControl(sstate->lower[2], HASH_SETPREFIX, 
	   (char *)args->credsp->keyblock.contents, args->credsp->keyblock.length);
  

  xTrace1(KERBSERVP, TR_EVENTS, "kerb_servOpen -- return session 0x%x",s);
  return s;
}

static xkern_return_t	
kerb_servOpenEnable( self, hlpRcv, hlpType, part)
     XObj self, hlpRcv, hlpType;
     Part *part;
{
  char buf[256];
  char idbuf[100];
  int  idlen, i, retval;
  PSTATE *ps = (PSTATE *)self->state;
  Enable *e;
  int buf_len=256;

  /* need to save a copy of the part list */
  if (partExternalize(part, buf, &buf_len) != XK_SUCCESS) {
    xTrace0(KERBSERVP, TR_ERRORS,
	    "kerb_servOpenEnable -- partExternalize failed!");
    return XK_FAILURE;
  }

  idlen = partToId(part, idbuf, 100);


  if (ps->paslen == 0) {
    /* this is the first openenable */
    ps->paslen = idlen;
    ps->pasMap = mapCreate(MAP_SZ, idlen);
  }

  if(ps->paslen != idlen) {
    xTrace0(KERBSERVP, TR_ERRORS, "kerb_servOpenEnable - bad idLen");
    return XK_FAILURE;
  } 
  
  if (mapResolve(ps->pasMap, idbuf, &e) != XK_FAILURE) {
    xTrace0(KERBSERVP, TR_ERRORS,
	    "kerb_servOpenEnable -- duplicate openenable");
    return XK_FAILURE;
  }

  for (i=0; i < 5; i++) {
    partInternalize(part, buf);
    retval = xOpenEnable(self, hlpType, xGetDown(self, i), part);
    if (retval != XK_SUCCESS) {
      xTrace1(KERBSERVP, TR_ERRORS,
	      "kerb_servOpenEnable -- could not openenable llp #%d", i);
      return XK_FAILURE;
    }
  }

  e = (Enable *)xMalloc(sizeof(Enable));
  if (!e) {
    xTrace0(KERBSERVP, TR_ERRORS,
	    "kerb_servOpenEnable -- malloc failed for enable object");
    return XK_FAILURE;
  }
  e->hlpRcv = hlpRcv;
  e->hlpType = hlpType;
  e->binding = mapBind(ps->pasMap, (char *)idbuf, (int) e);

  if (e->binding == ERR_BIND) {
    xTrace0(KERBSERVP, TR_ERRORS,
	    "kerb_servOpenEnable -- could not bind enable object");
    xFree((void *)e);
    return XK_FAILURE;
  }

  xTrace0(KERBSERVP, TR_ERRORS,"kerb_servOpenEnable -- exiting");
  return XK_SUCCESS;
  
}

static xkern_return_t	
kerb_servOpenDone( self, lls, llp, hlpType )
     XObj self, lls, llp, hlpType;
{
  PSTATE *ps = (PSTATE *)self->state;
  char buf[256];
  char idbuf[100];
  int  idlen, i;
  Part p[2];
  Enable *e;
  XObj s;
  SSTATE *ss;
  xkern_return_t val;
  int keylen=32;
  char keybuf[32];
  int retval;

  if (xControl(lls, GETPARTICIPANTS, buf, 256) < 1) {
    xTrace0(KERBSERVP, TR_ERRORS, "kerb_servOpenDone - getparticpants fails");
    return XK_FAILURE;
  }
  
  partInternalize(p, buf);
  idlen = part2ToId(p, idbuf, 100);

  if ( (idlen != ps->paslen) || 
       (mapResolve(ps->pasMap, idbuf, &e) != XK_SUCCESS) ) {
    xTrace0(KERBSERVP, TR_ERRORS, "kerb_servOpenDone -- no passive open");
    return XK_FAILURE;
  }

  xDuplicate(lls);

  idlen = partToId(p, idbuf, 100);
  if (ps->taglen == 0) {
    /* first opendone */
    ps->taglen = idlen;
    ps->tagMap = mapCreate(MAP_SZ, idlen);
  }

  if (ps->taglen != idlen) {
    xTrace0(KERBSERVP,TR_ERRORS,
	    "kerb_servOpenDone -- idlen differs from ps->taglen");
    return XK_FAILURE;
  }

  if (mapResolve(ps->tagMap, idbuf, &s) != XK_SUCCESS) {
    /* first opendone for this connection */
    s = xCreateSessn(sessn_init, e->hlpRcv, e->hlpType, self, 0,0);
    ss = (SSTATE *)xMalloc(sizeof(SSTATE));
    if (!ss) {
      xTrace0(KERBSERVP, TR_ERRORS,
	      "kerb_servOpenDone -- malloc failed for SSTATE");
      return XK_FAILURE;
    }
    bzero((char*)ss, sizeof(SSTATE));
    s->state = (void *)ss;
    ss->enhancement = None;
    ss->lower[0] = lls;
    ss->curr = 1;
    ss->tag_bind = mapBind(ps->tagMap, (char *)idbuf, (int)s);
    return XK_SUCCESS;
  }

  ss = (SSTATE *)s->state;
  ss->lower[ss->curr++] = lls;

  if (ss->curr == 5) {
    /* received all the opendones */
    mapRemoveBinding(ps->tagMap, ss->tag_bind);
    for(i=0;i<5;i++) {
      mapBind(ps->activemap, (char*)&ss->lower[i], (int)s);
    }
    val = xOpenDone(e->hlpRcv, s, self);
    if (val != XK_SUCCESS) {
      xTrace0(KERBSERVP, TR_ERRORS, "kerb_servOpenDone -- could not opendone hlp");
      for(i=0;i<5;i++) {
	xClose(ss->lower[i]);
      }
      return XK_FAILURE;
    }

    /* get the key from recvauth and pass it to the safe and priv paths */
    keylen = xControl(ss->lower[4], OBTAINKEY,keybuf, keylen);
    if (keylen <= 0) {
      xTrace0(KERBSERVP, TR_ERRORS,
	      "kerb_servOpenDone -- could not obtain key from Recvauth");
      for(i=0;i<5;i++) {
	xClose(ss->lower[i]);
      }
      bzero(keybuf, 32);
      return XK_FAILURE;
    }
    retval = xControl(ss->lower[3], CRYPT_IN_SET, keybuf,keylen);
    if (retval == -1) {
      xTrace0(KERBSERVP, TR_ERRORS, "CRYPT_IN_SET failed for session key");
      bzero(keybuf, 32);
      return XK_FAILURE;
    }

    retval = xControl(ss->lower[3], CRYPT_OUT_SET,keybuf, keylen);
    if (retval == -1) {
      xTrace0(KERBSERVP, TR_ERRORS, "CRYPT_OUT_SET failed for session key");
      bzero(keybuf, 32);
      return XK_FAILURE;
    }
    retval = xControl(ss->lower[2], HASH_SETPREFIX, keybuf, keylen);
    if (retval == -1) {
      xTrace0(KERBSERVP, TR_ERRORS, "HASH_SETPREFIX failed for session key");
      bzero(keybuf, 32);
      return XK_FAILURE;
    }
    bzero(keybuf, 32);
  }
  
  return XK_SUCCESS;
}



xkern_return_t kerb_servDemux(self, lls, m)
     XObj self, lls;
     Msg *m;
{
  XObj sessn;

  xTrace0(KERBSERVP, TR_EVENTS, "kerb_servDemux -- entered");

  if (mapResolve( ((PSTATE *)self->state)->activemap, &lls, &sessn) == XK_SUCCESS) {
    return xPop(sessn, lls, m, 0);
  } else {
    xTrace0(KERBSERVP, TR_SOFT_ERRORS, "lls not found in activemap");
    return XK_FAILURE;
  }
}


int kerb_servProtControl(self, op, buf, len)
     XObj self;
     int op, len;
     char *buf;
{
  return xControl(xGetDown(self, 0), op, buf, len);
}

int kerb_servSessnControl(self, op, buf, len)
     XObj self;
     int op, len;
     char *buf;
{
  int  enhance;
  
  
  switch(op) {
  case ENHANCEMENT: {
    if (len == sizeof(int)) {
      enhance = *(int*)buf;
      if (enhance == None || enhance == Safe || enhance == Private) {
	((SSTATE *)self->state)->enhancement = enhance;
	return len;
      } else {
	xTrace1(KERBSERVP, TR_SOFT_ERRORS, "bad enhancement option: %d",
		enhance);
	return -1;
      }
    } else {
      xTrace1(KERBSERVP, TR_SOFT_ERRORS, "bad enhancement len: %d", len);
      return -1;
    }
  }
  default: 
    return xControl( ((SSTATE*)self->state)->lower[0], op, buf, len);
  }
}


void sessn_init(self)
     XObj self;
{
  self->pop = kerb_servPop;
  self->push = kerb_servPush;
  self->control = kerb_servSessnControl;
}


xmsg_handle_t kerb_servPush(self, m)
     XObj self;
     Msg  *m;
{
  SSTATE        *ss = (SSTATE *)self->state;
  xmsg_handle_t  han;
  int            len;


  switch (ss->enhancement) {
  case None:
    {
      len = msgLen(m);
      msgPush(m, mycopypush, (char*)&len, sizeof(int),0);
      han = xPush(ss->lower[1], m);
      break;
    }
  case Safe:
    {
      han = xPush(ss->lower[2], m);
      break;
    }
  case Private:
    {
      han = xPush(ss->lower[3], m);
      break;
    }
  default:
    {
      xTrace1(KERBSERVP, TR_ERRORS, "Error -- current enhancement not supported: %d", ss->enhancement);
      return XMSG_ERR_HANDLE;
    }
  }
  xControl(ss->lower[0], JOINDONE, 0,0);
  return han;
}

xkern_return_t kerb_servPop(self, lls, m, hdr)
     XObj self, lls;
     Msg  *m;
     void *hdr;
{
  return xDemux(self, m);
}



static int
partToId(p, buf, maxLen)
    Part *p;
    char *buf;
    int maxLen;
{
    int i,j,len;
    char *ptr;
    int numPart = partLen(p);
    u_char *ptr2;
    
    xTrace0(KERBSERVP, TR_EVENTS, "partToLen - entered");
    len = 0;
    ptr = (char *)buf;
    for (j=0;j<numPart; j++) {
    	for (i=p[j].stack.top; i>0; i--)  {
       		len += p[j].stack.arr[i-1].len;
		if (len > maxLen) return 0;
        	bcopy((char *)p[j].stack.arr[i-1].ptr,(char *)ptr, p[j].stack.arr[i-1].len);
        	ptr += p[j].stack.arr[i-1].len;
    	}
    }

    xTrace1(KERBSERVP, TR_EVENTS, "partToLen - exiting len %d",len);
    xIfTrace(KERBSERVP, TR_EVENTS) {
        printf("part2ToLen - id= (");
        ptr2 = (u_char *)buf;
        for (i=0;i<len;i++) printf("%x",ptr2[i]);
        printf(")\n");
    }

    return(len);
}

static int
part2ToId(p, buf, maxLen)
    Part *p;
    char *buf;
    int maxLen;
{
    int i,len;
    char *ptr;
    u_char *ptr2;
    
    xTrace0(KERBSERVP, TR_EVENTS, "part2ToLen - entered");
    len = 0;
    ptr = (char *)buf;
    for (i=p[0].stack.top; i>0; i--)  {
       	len += p[1].stack.arr[i-1].len;
	if (len > maxLen) return 0;
       	bcopy((char *)p[1].stack.arr[i-1].ptr,(char *)ptr, p[1].stack.arr[i-1].len);
       	ptr += p[1].stack.arr[i-1].len;
    }
    xTrace1(KERBSERVP, TR_EVENTS, "part2ToLen - exiting len %d",len);
    xIfTrace(KERBSERVP, TR_EVENTS) {
        printf("part2ToLen - id= (");
        ptr2 = (u_char *)buf;
        for (i=0;i<len;i++) printf("%x",ptr2[i]);
        printf(")\n");
    }

    return(len);
}

static void mycopypush(src, dst, count, arg)
     void *src;
     char *dst;
     long  count;
     void *arg;
{
  bcopy((char *)src, dst, count);
}
