/*
 * photuris_i.h
 *
 * x-kernel v3.2
 *
 *
 *
 */

#define	PHOTURIS_ACTIVE_MAP_SZ	32
#define	PHOTURIS_ALGO_MAP_SZ	32


#define COOKIE_REQUEST				0
#define COOKIE_RESPOND				1
#define KEY_REQUEST				2
#define KEY_RESPOND				3
#define SIGNATURE_MESSAGE			4
#define KEY_CHANGE				5
#define ERROR_MESSAGE				7

#define XOPEN					-1
#define TIMEOUT					-2
#define GLOBAL_TIMEOUT				-3 


#define DELETED_STATE				-1
#define IDLE_STATE				0

#define INITIATOR_COOKIE_REQUEST  		100
#define INITIATOR_KEY_REQUEST			102
#define INITIATOR_SIGNATURE_MESSAGE		104
#define INITIATOR_SIGNATURE_MESSAGE_FINISHED	105

/* The first time the responder has state is if a key_request is received */

#define RESPONDER_KEY_RESPOND			129
#define RESPONDER_SIGNATURE_MESSAGE		130

#define MAX_ATTRIBUTE_LIST	5
#define MAX_GROUP_LIST		3
#define MAX_ATTRIBUTE_VALUE	256
#define COOKIE_SIZE		16
#define MAX_SIGNATURE_SIZE	32
#define MAX_CERTIFICATE_SIZE	50
#define MAX_VERIFICATION_SIZE	256
#define MAX_PUBLIC_VALUE_SIZE	512
#define MAX_KEY_SIZE		512
						

typedef int tSPI;

typedef struct {
char	value[COOKIE_SIZE];
} Cookie;

typedef struct {
unsigned char	type;
unsigned char	length;
char 	value[MAX_ATTRIBUTE_VALUE];
} Attribute;

typedef struct {
int length;
Attribute  list[MAX_ATTRIBUTE_LIST];
} Attribute_list;

typedef struct {
int index;
} Group;

typedef struct {
int length;
Group list[MAX_GROUP_LIST];
} Groupe_list;

typedef struct {
int length;
char value[MAX_SIGNATURE_SIZE];
} Signature;

typedef struct {
int length;
char value[MAX_CERTIFICATE_SIZE];
} Certificate;

typedef struct {
int length;
char value[MAX_VERIFICATION_SIZE];
} Verification;

typedef struct {
int length;
char value[MAX_PUBLIC_VALUE_SIZE];
} Public_value;

typedef struct {
int length;
char value[MAX_KEY_SIZE];
} Key;

typedef struct {
int	type;
} Cookie_Request;

typedef struct {
int 	type;
Groupe_list Offered_Schemes;
} Cookie_Response;

typedef struct {
int 	type;
Group	Scheme_Choice;
Public_value	Initiator_Exchange_Value;
Attribute_list Initiator_Offered_Attributes;
} Key_Request;

typedef struct {
int 	type;
Public_value	Responder_Exchange_Value;
Attribute_list Responder_Offered_Attributes;
} Key_Response;

typedef struct {
int 	type;
int 	Life_Time;
tSPI	Security_Parameter_Index;
Attribute Signature_Choice;
Signature Signature;
Certificate Certificate;
Attribute_list Attribute_Choices;
} Signature_Message;

typedef struct {
int 	type;
int	Life_Time;
tSPI	Security_Parameter_Index;
Attribute Validity_Choice;
Verification Verification;
Attribute_list Attribute_Choices;
} Key_Change;

typedef struct {
int	type;
int	code;
Attribute_list Attributes_Needed;
} Error_Message;

typedef struct {
	char 		user[MAX_CERTIFICATE_SIZE];
        Map		actMap;
	Map		algoMap;
	XObj		group_protocols[MAX_GROUP_LIST];
	XObj		group_sessions[MAX_GROUP_LIST];
	int		group_timeout;
	int		group_timeout_ini;
	XObj		attribute_protocols[MAX_ATTRIBUTE_LIST];
	XObj		km;
	XObj		kmid;
	Groupe_list	supported_groups;
	Attribute_list	supported_attributes;
	int		default_key_lifetime;
	Event		timeout;
	int		timeout_time;
	int		global_timeout;
} PState;


typedef struct {
	char 		user[MAX_CERTIFICATE_SIZE];
	long		local_port;
	int		global_timeout;
	int		timeout;
	Semaphore	initial_exchange;
	Group		Group_index;
	Groupe_list	Offered_Schemes;
	Public_value	gx;
	XObj		gx_session;
	Public_value	gy;
	Public_value	gxy;
	int		cookie_random_value;
	Attribute	I_Signature_Choice;
	Attribute	R_Signature_Choice;
	Attribute_list	I_Offered_Attributes;
	Attribute_list	R_Offered_Attributes;
	Attribute_list	I_Attribute_Choices;
	Attribute_list	R_Attribute_Choices;
	Cookie		I_Cookie;
	Cookie		R_Cookie;
	tSPI		I_SPI;
	int		I_lifetime; 
	tSPI		R_SPI;		
	int		R_lifetime;
	XIpsecMode	mode;
	int		state;
	Semaphore	process;
	Key_Request *	incomming_key_message;
	Key_Response *	outgoing_key_message;
	Signature_Message * incomming_signature_message;
	Signature_Message * outgoing_signature_message;
	IPhost		myhost;
	IPhost		peerhost;
} SState;


