/*
 * $RCSfile: ppp_options.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1994  Arizona Board of Regents
 *
 * $Log: ppp_options.c,v $
 * Revision 1.2  1994/10/27 17:27:28  andrey
 * removed non-ansi definitions
 *
 * Revision 1.1  1994/10/24  20:44:34  andrey
 * Initial revision
 *
 *
 * $Revision: 1.2 $
 * $Date: 1994/10/27 17:27:28 $
 */

/*
 * common options list handling routines for LCP and IPCP
 *
 * exports list handlers
 *
 */

void ppp_add_option(PPP_options * opts, void * op, u_char type);

void ppp_options_append(PPP_options * to, PPP_options * from);

void ppp_copy_options(PPP_options * to, PPP_options * from,
  			void * (*duplicate_option)(void * ptr, u_char type));

void ppp_destroy_options(PPP_options * opt,
                         void (*destroy_opt)(void * ptr, u_char type));

int ppp_options_match(PPP_options * from, PPP_options * to,
                     int (*option_matches)(void * fr, void * t, u_char type));

PPP_options * ppp_option_differences(PPP_options * from, PPP_options * sub,
  int (*option_matches)(void * fr, void * t, u_char ty),
  void * (*duplicate_option)(void * ptr, u_char ty));

/* -------------------------------------------------------------------- */
/*                     options processing                               */
/* -------------------------------------------------------------------- */

static void
append_ole( PPP_options * opts, opt_list_elem * ole_ptr)
{
  opt_list_elem * olep;

  if (olep = opts->first_opt) {
    while (olep->next)
      olep = olep->next;
    link_ole(olep, ole_ptr);
  }
  else {
    opts->first_opt = ole_ptr;
  }
}

static opt_list_elem *
append_new(PPP_options * opts)
{
  opt_list_elem * t = X_NEW(opt_list_elem);
  opt_list_elem_init(t);

  append_ole(opts,t);
  return t;
}

void
ppp_add_option( PPP_options * opts, void * op, u_char type)
{
  opt_list_elem * t = append_new(opts);
  t->type = type;
  t->opt_ptr = op;
}

/*
 * move from's option list to to, and erase from's pointer
 */
void
ppp_options_append(PPP_options * to, PPP_options * from)
{
  opt_list_elem * last_opt;

  if (!first_option(from)) return;
  last_opt = first_option(to);
  if (last_opt) {
    while(last_opt->next)
      last_opt = next_option(last_opt);
    link_ole(last_opt, first_option(from));
  }
  else
    from->first_opt = first_option(from);

  from->first_opt = 0;
}

void
ppp_copy_options( PPP_options * to, PPP_options * from,
  void * (*duplicate_option)(void * ptr, u_char type))
{
  opt_list_elem * ole;

  for(ole = first_option(from); ole; ole = next_option(ole)) {
    ppp_add_option(to, duplicate_option(ole->opt_ptr, ole->type), ole->type);
  }
}


static void
destroy_list_elem(opt_list_elem * ole,
  void (*destroy_opt)(void * ptr, u_char type) )
{
  destroy_opt(ole->opt_ptr, ole->type);
  FREE(ole);
}

void
ppp_destroy_options(PPP_options * opt,
  void (*destroy_opt)(void * ptr, u_char type))
{
  opt_list_elem * ole, * olef;
  ole = first_option(opt);
  while (ole) {
     olef = next_option(ole);
     destroy_list_elem(ole, destroy_opt);
     ole = olef;
  }
}

void
ppp_print_options(PPP_options * opt,
  void (*print_opt)(void * ptr, u_char type))
{
  opt_list_elem * ole, * olef;
  ole = first_option(opt);
  while (ole) {
     olef = next_option(ole);
     print_opt(ole->opt_ptr, ole->type);
     ole = olef;
  }
}

int
ppp_options_match(PPP_options * from, PPP_options * to,
  int (*option_matches)(void * fr, void * t, u_char type))
{
  opt_list_elem * olef = 0, * olet = 0;

  olef = first_option(from);
  olet = first_option(to);
  while (olef) {
    u_char type = olef->type;
    if (!olet) return 0;
    if (!olet->type != type) return 0;
    if (!option_matches(olef->opt_ptr, olet->opt_ptr, type)) return 0;
    olef = next_option(olef);
    olet = next_option(olet);
  }

  if (olet) return 0;
  return 1;
}

PPP_options *
ppp_option_differences(PPP_options * from, PPP_options * sub,
  int (*option_matches)(void * fr, void * to, u_char type),
  void * (*duplicate_option)(void * ptr, u_char type))
{
  PPP_options * diff = 0;
  opt_list_elem * ole, * ole2;

  for (ole = first_option(from); ole; ole = next_option(ole)) {
    u_char t = ole->type;
    int found = 0;
    for (ole2 = first_option(sub); ole2; ole2 = next_option(ole2)) {
      u_char t2 = ole2->type;
      if (t2 == t) { found = 1; break; }
    }
    if (found && option_matches(ole->opt_ptr, ole2->opt_ptr, t))
      continue;
    else {
      if (!diff) diff = X_NEW(PPP_options);
      ppp_add_option(diff,
                     duplicate_option(ole->opt_ptr, ole->type), ole->type);
    }
  }
  return diff;
}


