/*
 * x-kernel v3.2
 *
 * Copyright (c) 1993  Arizona Board of Regents
 */

#include "xkernel.h"
#include "ripgrd_i.h"
#include "ripgrd.h"
#include "km.h"

#ifdef __STDC__

static void		getproc_protl( XObj );
static void		getproc_sessn( XObj );
static XObj		ripgrdCreateSessn( XObj, XObj, XObj, XObj, Part * );
static XObj		ripgrdOpen( XObj, XObj, XObj, Part * );
static xkern_return_t	ripgrdOpenDisable( XObj, XObj, XObj, Part * );
static xkern_return_t	ripgrdOpenDone( XObj, XObj, XObj, XObj );
static xkern_return_t	ripgrdOpenEnable( XObj, XObj, XObj, Part * );
static int		ripgrdControlProtl( XObj, int, char *, int );
static int		ripgrdControlSessn( XObj, int, char *, int );
static xkern_return_t	ripgrdClose( XObj );
static xkern_return_t	ripgrdPop( XObj, XObj, Msg *, void * );
static xmsg_handle_t	ripgrdPush( XObj, Msg * );
static xkern_return_t	ripgrdProtlDemux( XObj, XObj, Msg * );
static xkern_return_t	ripgrdSessnDemux( XObj, XObj, Msg * );
static long             getRipCmd( void *, char *, long, void * );
static xkern_return_t   romgrpid(XObj, char **, int, int, void * );

#else

static void		getproc_protl();
static void		getproc_sessn();
static XObj		ripgrdCreateSessn();
static XObj		ripgrdOpen();
static xkern_return_t	ripgrdOpenDisable();
static xkern_return_t	ripgrdOpenDone();
static xkern_return_t	ripgrdOpenEnable();

#endif /* __STDC__ */


int traceripgrdp;

static XObjRomOpt rmopts[] = {
                               { "grpid", 3, romgrpid },
                             };



void
ripgrd_init( self )
    XObj 	self;
{
    PState	*ps;
    
    xTraceP0(self, TR_GROSS_EVENTS, "init");
    xAssert(xIsProtocol(self));
    if ( (! xIsProtocol(xGetDown(self, 0))) || (!xIsProtocol(xGetDown(self, 2))) ) {
	xError("RIPGRD down vector is misconfigured");
	return;
    }
    ps = X_NEW(PState);
    self->state = (VOID *)ps;
    getproc_protl(self);
    ps->activeMap = mapCreate(RIPGRD_ACT_MAP_SZ, sizeof(XObj));
    ps->passiveMap = mapCreate(RIPGRD_PAS_MAP_SZ, sizeof(XObj));
    findXObjRomOpts(self, rmopts, sizeof(rmopts)/sizeof(XObjRomOpt), 0);
}


static XObj
ripgrdOpen( self, hlpRcv, hlpType, p )
    XObj	self, hlpRcv, hlpType;
    Part 	*p;
{
    XObj	s, lls;
    PState	*ps = (PState *)self->state;
    Part	part[2];
    
    bcopy ((char *)p, (char *)part, sizeof(part));
    xTraceP0(self, TR_MAJOR_EVENTS, "open");
    if ( (lls = xOpen(self, hlpType, xGetDown(self, 0), p)) == ERR_XOBJ ) {
	xTraceP0(self, TR_ERRORS, "open: could not open lower session");
	return ERR_XOBJ;
    }
    if ( mapResolve(ps->activeMap, &lls, &s) == XK_SUCCESS ) {
	xTraceP0(self, TR_MORE_EVENTS, "open -- found an existing one");
    } else {
	xTraceP0(self, TR_MAJOR_EVENTS, "open -- creating a new one");
	if ( (s = ripgrdCreateSessn(self, hlpRcv, hlpType, lls, part)) == ERR_XOBJ ) {
	    xClose(lls);
	}
    }
    return s;
}


static XObj
ripgrdCreateSessn( self, hlpRcv, hlpType, lls, part )
    XObj 	self, hlpRcv, hlpType, lls; Part *part;
{
    XObj	s;
    PState	*ps = (PState *)self->state;
    XObj	sigs = ERR_XOBJ;
    XObj	SessnArray[2];
    
    if ( (sigs = xOpen(self, hlpType, xGetDown(self, 2), part)) == ERR_XOBJ ) {
	xTraceP0(self, TR_ERRORS, "open: could not open lower session");
	return ERR_XOBJ;
    }
    SessnArray[0] = lls;
    SessnArray[1] = sigs;

    if ((s = xCreateSessn(getproc_sessn, hlpRcv, hlpType, self, 2, SessnArray))
			== ERR_XOBJ ) {
	xTraceP0(self, TR_ERRORS, "xCreateSessn fails");
	return ERR_XOBJ;
    }
    s->binding = mapBind(ps->activeMap, &lls, s);
    if ( s->binding == ERR_BIND ) {
	xTraceP0(self, TR_ERRORS, "mapBind fails in createSessn");
	return ERR_XOBJ;
    }
    /*
     * The lower sessions' up field is made to point to this
     * ripgrd session (not the protocol)
     */
    xSetUp(lls, s);
    xTraceP1(self, TR_MAJOR_EVENTS, "open returns %x", s);
    return s;
}


static int
ripgrdControlSessn( s, opcode, buf, len )
    XObj 	s;
    int 	opcode, len;
    char 	*buf;
{
    xTraceP0(s, TR_EVENTS, "controlsessn");
    switch ( opcode ) {
      default:
	/*
	 * All other opcodes are forwarded to the lower session.  
	 */
	return xControl(xGetDown(s, 0), opcode, buf, len);
    }
}


static int
ripgrdControlProtl( self, opcode, buf, len )
    XObj 	self;
    int 	opcode, len;
    char 	*buf;
{
    /*
     * All opcodes are forwarded to the lower protocol
     */
    return xControl(xGetDown(self, 0), opcode, buf, len);
}


static xkern_return_t
ripgrdOpenEnable( self, hlpRcv, hlpType, p )
    XObj 	self, hlpRcv, hlpType;
    Part 	*p;
{
    PState	*ps = (PState *)self->state;
    
    xTraceP0(self, TR_MAJOR_EVENTS, "open enable");
    return defaultVirtualOpenEnable(self, ps->passiveMap, hlpRcv, hlpType,
				    self->down, p);
}


static xkern_return_t
ripgrdOpenDisable( self, hlpRcv, hlpType, p )
    XObj 	self, hlpRcv, hlpType;
    Part 	*p;
{
    PState	*ps = (PState *)self->state;

    return defaultVirtualOpenDisable(self, ps->passiveMap, hlpRcv, hlpType,
				     self->down, p);
}


static xkern_return_t
ripgrdClose( s )
    XObj 	s;
{
    PState		*ps;
    xkern_return_t	res;
    XObj		lls;
    
    xTraceP1(s, TR_MAJOR_EVENTS, "close of session %lx", (u_long)s);
    xAssert(s->rcnt == 0);
    ps = (PState *)xMyProtl(s)->state;
    res = mapRemoveBinding(ps->activeMap, s->binding);
    xAssert( res != XK_FAILURE );
    lls = xGetDown(s, 0);
    xAssert(xIsSession(lls));
    xSetUp(lls, xMyProtl(s));
    xClose(lls);
    xClose(xGetDown(s, 1));
    xDestroy(s);
    return XK_SUCCESS;
}


static xmsg_handle_t
ripgrdPush( s, msg )
    XObj	s;
    Msg 	*msg;
{
    xTraceP0(s, TR_EVENTS, "push");
    xPush(xGetDown(s, 1), msg);
    return xPush(xGetDown(s, 0), msg);
}


static xkern_return_t
ripgrdOpenDone( self, lls, llp, hlpType )
    XObj	self, lls, llp, hlpType;
{
    XObj	s;
    PState	*ps = (PState *)self->state;
    Enable	*e;
    Part 	p[2];
    char 	buffer[2048];
    
    xTraceP0(self, TR_MAJOR_EVENTS, "In openDone");
    if ( self == hlpType ) {
	xTraceP0(self, TR_ERRORS, "self == hlpType in openDone");
	return XK_FAILURE;
    }
    /*
     * check for openEnable
     */
    
    if ( mapResolve(ps->passiveMap, &hlpType, &e) == XK_FAILURE ) {
	/* 
	 * This shouldn't happen
	 */
	xTraceP0(self, TR_ERRORS,
		"openDone: Couldn't find hlp for incoming session");
	return XK_FAILURE;
    }
    xControl(lls,GETPARTICIPANTS,buffer,sizeof(buffer));
    partInternalize(p,buffer);
    if ( (s = ripgrdCreateSessn(self, e->hlpRcv, e->hlpType, lls, p))
							== ERR_XOBJ ) {
	return XK_FAILURE;
    }
    xDuplicate(lls);
    xTraceP0(self, TR_EVENTS,
	    "passively opened session successfully created");
    return xOpenDone(e->hlpRcv, s, self);
}


static xkern_return_t
ripgrdProtlDemux( self, lls, m )
    XObj	self, lls;
    Msg		*m;
{
    xTraceP0(self, TR_ERRORS, "protlDemux called!!");
    return XK_SUCCESS;
}


/* 
 * ripgrdPop and ripgrdSessnDemux must be used (i.e., they can't be
 * bypassed) for the UPI reference count mechanism to work properly. 
 */
static xkern_return_t
ripgrdPop( s, lls, msg, h )
    XObj 	s, lls;
    Msg 	*msg;
    VOID	*h;
{
    char        cmd;
    XObj 	kms;
    Part 	p[2], grpid;
    IPhost 	kmcontents[GRPMEM_MAX_PARTS];
    IPhost      null = {0, 0, 0, 0};
    int		i=0;
    char	buffer[2048];
    IPhost	*gwHost;
    long	*gwPort;
    XObj 	myprotl = s->myprotl;

    
    msgPop (msg, getRipCmd, &cmd, msgLen(msg), 0);
    xTraceP0(s, TR_EVENTS, "pop");
    if (cmd != RIP_RESPONSE) 
        return xDemux(s, msg);
    else {
        partInit(&grpid,1);
        partPush( grpid, &((PState *)myprotl->state)->grpid, sizeof(IDTYPE) );

        kms = xOpen( myprotl, myprotl, xGetDown(myprotl,KM_PROTL), &grpid );
        if ( kms == ERR_XOBJ ) {
            xError("grpmem: unable to open key manager session");
        }
        xControl(s,GETPARTICIPANTS,buffer,sizeof(buffer));
        partInternalize(p,buffer);
        gwPort = (long *)partPop(p[0]);
        gwHost = (IPhost *)partPop(p[0]);
        xControl(kms,KM_RESOLVE, (void*) kmcontents, sizeof(kmcontents) );
        while (!IP_EQUAL(kmcontents[i], null)) {
	    if (IP_EQUAL(*gwHost, kmcontents[i])) {
    		xTraceP1(s, TR_EVENTS, "accepting RIP update from %s", ipHostStr(gwHost));
		return xDemux(s, msg);
	    }
            i++;
        }
    }
    xTraceP1(s, TR_EVENTS, "rejecting RIP update from %s", ipHostStr(gwHost));
    return XK_SUCCESS;
}

static long
getRipCmd (cmd, src, len, arg)
    void 	*cmd;
    char 	*src;
    long 	len;
    void 	*arg;
{
    char *ripcmd = cmd;

    if (len  >= RIPHSIZE)
        *ripcmd = ((struct rip *)src)->rip_cmd;
    return 0;
}


static xkern_return_t
ripgrdSessnDemux( self, lls, msg )
    XObj 	self, lls;
    Msg 	*msg;
{
    xTraceP0(self, TR_EVENTS, "session demux");
    return xPop(self, lls, msg, 0);
}


static void
getproc_sessn( s )
    XObj 	s;
{
    xAssert(xIsSession(s));
    
    s->push = ripgrdPush;
    s->pop = ripgrdPop;
    s->close = ripgrdClose;
    s->control = ripgrdControlSessn;
    /* 
     * RIPGRD sessions will look like a protocol to lower sessions, so we
     * need a demux function
     */
    s->demux = ripgrdSessnDemux;
}


static void
getproc_protl(p)
    XObj p;
{
    xAssert(xIsProtocol(p));

    p->control = ripgrdControlProtl;
    p->open = ripgrdOpen;
    p->openenable = ripgrdOpenEnable;
    p->opendisable = ripgrdOpenDisable;
    p->demux = ripgrdProtlDemux;
    p->opendone = ripgrdOpenDone;
}

static xkern_return_t
romgrpid(self, ptr, fields, line, arg)
  XObj self;
  char **ptr;
  int  fields, line;
  void *arg;
{
  PState* ps = (PState*) self->state;
  xTraceP0(self, TR_MAJOR_EVENTS, "grpid rom opt");
  return sscanf(ptr[2],"%d",&ps->grpid)<1? XK_FAILURE: XK_SUCCESS;
}
 

