
#include "xkernel.h"
#include "tcp.h"
#include "tcp_forward_i.h"

int	tracetcp_forwardp;


static XObj		tcp_forwardCreateSessn( XObj, XObj, XObj, ActiveId * );
static xkern_return_t	tcp_forwardDemux( XObj, XObj, Msg * );
static xkern_return_t	tcp_forwardCloseSessn( XObj );
static xkern_return_t	tcp_forwardCloseDone( XObj );
static void		getproc_sessn(XObj);
void			tcp_forward_init(XObj);

static xkern_return_t romtarget_address(XObj, char **, int, int, void *);
static xkern_return_t romport(XObj, char **, int, int, void *);

static XObjRomOpt rmopts[2] = {{ "target_address", 2, romtarget_address},
			       { "port", 2, romport}	};

static xkern_return_t
romtarget_address(self, ptr, fields, line, arg)
XObj self;
char **ptr;
int  fields, line;
void *arg;
{
  PSTATE     *pstate = (PSTATE *)(self->state);

  xTrace1(tcp_forwardp, TR_ALWAYS, "TCP_FORWARD set IP address %s", ptr[2]);
  return str2ipHost((IPhost*) &(pstate->forward_to), ptr[2]);
}

static xkern_return_t
romport(self, ptr, fields, line, arg)
XObj self;
char **ptr;
int  fields, line;
void *arg;
{
/*
ftp-data	20/tcp
ftp		21/tcp
telnet		23/tcp
www		80/tcp
*/
  Part	part[3];
  long	localPort;
  PSTATE     *pstate = (PSTATE *)(self->state);

  xTrace1(tcp_forwardp, TR_ALWAYS, "TCP_FORWARD openenable port  %s", ptr[2]);
  localPort = atoi(ptr[2]);
  partInit(part, 1);
  partPush(part[0], ANY_HOST, 0);
  partPush(part[0], &localPort, sizeof(long));
  if ( xOpenEnable(self, self, xGetDown(self, 0), part) == XK_FAILURE ) {
	xTrace0(tcp_forwardp, TR_ALWAYS,
		"tcp_forward_init: can't openenable transport protocol");
	xFree((char *) pstate);
	return XK_FAILURE;
   }  
   return XK_SUCCESS;
}

/*
 * tcp_forward_init
 */
void
tcp_forward_init(self)
    XObj self;
{
    PSTATE	*pstate;
    XObj	llp;
    XTime	t;

    xTrace0(tcp_forwardp, TR_GROSS_EVENTS, "TCP_FORWARD init");
    xGetTime(&t);
    srand(t.sec);
    pstate = X_NEW(PSTATE);
    bzero((char *)pstate, sizeof(PSTATE));
    self->state = (VOID *)pstate;
    self->demux = tcp_forwardDemux;
    self->closedone = tcp_forwardCloseDone;
    if ( ! xIsProtocol(llp = xGetDown(self, 0)) ) {
	Kabort("TELNET could not get lower protocol");
    }
    xControl(llp,GETMYHOST,(char *)&(pstate->my_address), sizeof(IPhost));
    pstate->activemap = mapCreate(ACTIVE_MAP_SIZE, sizeof(ActiveId));
    semInit(&pstate->demux,1);
    findXObjRomOpts(self, rmopts, 2, 0);
}

static XObj
tcp_forwardCreateSessn(self, hlpRcv, hlpType, key)
    XObj self, hlpRcv, hlpType;
    ActiveId *key;
{
    XObj	s;
    SSTATE	*sstate;
    PSTATE	*pstate;

    pstate = (PSTATE *)self->state;
    s = xCreateSessn(getproc_sessn, hlpRcv, hlpType, self, 1, &key->lls);
    s->binding = mapBind(pstate->activemap, key, s);
    sstate = X_NEW(SSTATE);
    s->state = sstate;
    return s;
}



/*
 * tcp_forwardDemux
 */
static xkern_return_t
tcp_forwardDemux(self, lls, dg)
    XObj self;
    XObj lls;
    Msg *dg;
{
    ActiveId  	activeid;
    PSTATE      *pstate;
    XObj	from_s;
    XObj	to_s;
    u_short	bufspace = 1024;

    pstate = (PSTATE *)self->state;
    xTrace0(tcp_forwardp, TR_EVENTS, "tcp_forward Demux");
    semWait(&pstate->demux);  
    activeid.lls = lls;
    if ( mapResolve(pstate->activemap, &activeid, &from_s) == XK_FAILURE ) 
    {
   	long	to_port;
/*   	long	local_port = ANY_HIGH_PORT;*/
   	long	local_port ;		 /* otherwise there are problems with preexisting
						connections */
    	char	buf[100];
   	Part	p[3];
	XObj	to_lls;
  	local_port = (rand() % 50000) + 4000;
	xTrace0(tcp_forwardp, TR_MAJOR_EVENTS, "Create new tcp_forward session");
	if ( xControl(lls, GETPARTICIPANTS, buf, 100) == XK_FAILURE)
	{
    		semSignal(&pstate->demux);  
		return XK_FAILURE;
	}
	partInit(p,3);
	partInternalize(p,buf);
	to_port = *((long *)partPop(p[LOCAL_PART]));
    	xTrace0(tcp_forwardp, TR_EVENTS, "tcp_forward Demux: got participants");
	partInit(p,2);
	partPush(p[REMOTE_PART],&pstate->forward_to,sizeof(IPhost));
	partPush(p[REMOTE_PART],&to_port,sizeof(long));
	partPush(p[LOCAL_PART],&pstate->my_address,sizeof(IPhost));
	partPush(p[LOCAL_PART],&local_port,sizeof(long));
    	xTrace1(tcp_forwardp, TR_EVENTS, "tcp_forward Demux: open connection to port %u",to_port);
	if ((to_lls = xOpen(self,self,xGetDown(self,0),p)) == ERR_XOBJ)
	{
    		xTrace0(tcp_forwardp, TR_EVENTS, "tcp_forward Demux: tcp open failed");
    		semSignal(&pstate->demux);  
		return XK_FAILURE;
	}
	xDuplicate(lls);
    	xTrace0(tcp_forwardp, TR_EVENTS, "tcp_forward Demux: Creat Sessions");
	from_s = tcp_forwardCreateSessn(self,self,self,&activeid);
    	activeid.lls = to_lls;
	to_s = tcp_forwardCreateSessn(self,self,self,&activeid);
	((SSTATE *)from_s->state)->to_session = to_s;
	((SSTATE *)to_s->state)->to_session = from_s;
    	xTrace0(tcp_forwardp, TR_EVENTS, "tcp_forward Demux: Session Created");
    }
    else
    {
	to_s = ((SSTATE *)from_s->state)->to_session;
    }
    xTrace2(tcp_forwardp, TR_EVENTS, "tcp_forward Demux: Push Message %x --> %x",from_s, to_s);
    xPush(xGetDown(to_s,0),dg);
    xControl(xGetDown(to_s,0), TCP_PUSH, NULL, 0);	
    xControl(lls, TCP_SETRCVBUFSPACE, (char *) &bufspace, sizeof(bufspace));	
    xTrace0(tcp_forwardp, TR_EVENTS, "tcp_forward Demux: Done");
    semSignal(&pstate->demux);  
    return XK_SUCCESS;
}


/*
 * tcp_forwardCloseSessn
 */
static xkern_return_t
tcp_forwardCloseSessn(s)
    XObj s;
{
    SSTATE	*sstate;
    PSTATE	*pstate;

    sstate = (SSTATE *)s->state;
    pstate = (PSTATE *)s->myprotl->state;

 
    xAssert(xIsSession(s));
    xTrace1(tcp_forwardp, TR_EVENTS, "TCP_FORWARD close of session %x", s);
    mapRemoveBinding(pstate->activemap,s->binding);
    if (sstate->to_session != NULL)
    {
	((SSTATE *)sstate->to_session->state)->to_session = NULL;
	xClose(sstate->to_session);
    }
    xClose(xGetDown(s,0));
    xFree((char*)sstate);
    xDestroy(s);
    return XK_SUCCESS;
}

/*
 * tcp_forwardCloseDone
 */
static xkern_return_t
tcp_forwardCloseDone(s)
    XObj s;
{
    XObj	ss;
    ActiveId  	activeid;

    activeid.lls = s;
    xTrace1(tcp_forwardp, TR_EVENTS, "TCP_FORWARD closeDone of session %x", s);

    if ( mapResolve(((PSTATE*)s->up->state)->activemap, &activeid, &ss) == XK_FAILURE ) 
    {
	return XK_SUCCESS; /* session not used anyway */
    }

    xAssert(xIsSession(ss));
    xClose(ss);

    return XK_SUCCESS;
}
static void
getproc_sessn(s)
    XObj s;
{
  xAssert(xIsSession(s));
  s->close = tcp_forwardCloseSessn;
}
