/*
 * $RCSfile: symbol_defs.h,v $
 *
 * Universal Stub Compiler
 *
 * Copyright (c) 1994    Arizona Board of Regents
 *
 * $Log: symbol_defs.h,v $
 * Revision 1.1  1994/10/13 17:43:30  dorgival
 * Initial revision
 *
 */

#ifndef _INSYMBOLDEFSH
#define _INSYMBOLDEFSH

typedef char *ByteOrder;
typedef signed char BitInfo;

typedef enum { 
  INT_CONST, 
  DOUBLE_CONST, 
  STRING_CONST
} ConstantType;

typedef enum { 
  NULL_TYPE = 0,
  CHAR_TYPE = 1, 
  SHORT_TYPE = 2, 
  INT_TYPE = 4, 
  LONG_TYPE = 8,
  SIGNED_TYPE = 16, 
  STRUCT_TYPE = 32, 
  ENUM_TYPE = 64, 
  UNION_TYPE = 128, 
  UNSIGNED_TYPE = 256
} SpecifierTypeField;

typedef enum { 
  BITFIELD_TYPE = 1, 
  ARRAY_TYPE = 2, 
  MODIFIER_TYPE = 4, 
  FUNCTION_TYPE = 8,
  POINTER_TYPE = 16
} DeclaratorTypeField;

typedef enum { 
  DECLARATOR, 
  SPECIFIER
} TypeClass;

typedef enum {
  NO_UNIT,
  BIT_UNIT,
  BYTE_UNIT
} ExprUnit;

typedef struct constant {
  ConstantType the_type;
  ExprUnit theunit;
  union {
    char *string_value;
    long int_value;
    double double_value;
  } value;
  struct constant *next;
} Constant;

#define CONST_TYPE(a)   ((a)->the_type)
#define CONST_STRING(a) ((a)->value.string_value)
#define CONST_INT(a)    ((a)->value.int_value)
#define CONST_DOUBLE(a) ((a)->value.double_value)
#define CONST_UNIT(a)   ((a)->theunit)

typedef struct structdef {
  char *name;
  int level;
  struct symbol *fields;
} StructDef;

typedef struct specifier {
  SpecifierTypeField thetype; 
  StructDef *struct_def; /* if it's a struct */
} Specifier;

typedef struct {
  int size;
  int memsize;     /* If this is a type modifier it is the TOTAL size of
			this type. If it is an ARRAY declarator it is the
			size of the elements of the array. Got it? */
  int alignment;
  ByteOrder byte_order;
  int offset;
  int complete_type;
} TypeData;

#define NULL_TYPEDATA         { 0, 0, 0, NULL, 0, 1}
#define TYPEDATA_SIZE(a)      ((a)->size)
#define TYPEDATA_MEMSIZE(a)   ((a)->memsize)
#define TYPEDATA_ALIGN(a)     ((a)->alignment)
#define TYPEDATA_BORDER(a)    ((a)->byte_order)
#define TYPEDATA_OFFSET(a)    ((a)->offset)
#define TYPEDATA_COMPLETE(a)  ((a)->complete_type)

typedef struct function_data {
  struct symbol* param_list;
  struct expr*   body;
} FunctionData;

typedef struct declarator {
  DeclaratorTypeField thetype;
  int num_ele;
  union {
    TypeData thedata;
    FunctionData function;
  } u;
  BitInfo *bitinfo;
  struct expr* var_array; /* ident expr with name of variable */
} Declarator;

typedef enum {
  USIT_NOT_SEEN = 0,
  USIT_SEEN     = 1
} UsitTypeSeen;

typedef struct type {
  TypeClass class;  /* declarator or specifier?     */
  char* type_name;  /* if so, it is the type name   */
  char tdef;        /* is it a typedef?             */
  char native;      /* is it a native type?         */
  char addnative;   /* was the "native" class used? */
  UsitTypeSeen usit_seen[2];
  union {
    Specifier s;
    Declarator d;
  } u;
  TypeData *type_modifiers;
  struct type *next, *etype;
  struct type *next_bro, *prev_bro;
} Type;

#define TYPE_CLASS(a)       ((a)->class)
#define TYPE_USIT_SEEN(a,b) ((a)->usit_seen[b])
#define TYPE_TDEF(a)        ((a)->tdef)
#define TYPE_NAME(a)        ((a)->type_name)
#define TYPE_NATIVE(a)      ((a)->native)
#define TYPE_ADDNATIVE(a)   ((a)->addnative)
#define TYPE_NEXT(a)        ((a)->next)
#define TYPE_ETYPE(a)       ((a)->etype)
#define TYPE_NEXTBRO(a)     ((a)->next_bro)
#define TYPE_PREVBRO(a)     ((a)->prev_bro)
#define TYPE_SPECIFIER(a)   ((a)->u.s)
#define TYPE_DECLARATOR(a)  ((a)->u.d)
#define TYPE_DECL_TYPE(a)   ((a)->u.d.thetype)
#define TYPE_DECL_NUMEL(a)    ((a)->u.d.num_ele)
#define TYPE_DECL_COMPLETE(a) ((a)->u.d.u.thedata.complete_type)
#define TYPE_DECL_SIZE(a)     ((a)->u.d.u.thedata.size)
#define TYPE_DECL_MEMSIZE(a)  ((a)->u.d.u.thedata.memsize)
#define TYPE_DECL_BITINFO(a)  ((a)->u.d.bitinfo)
#define TYPE_DECL_VAR_ARRAY(a)  ((a)->u.d.var_array)
#define TYPE_DECL_ALIGN(a)    ((a)->u.d.u.thedata.alignment)
#define TYPE_DECL_BORDER(a)   ((a)->u.d.u.thedata.byte_order)
#define TYPE_DECL_OFFSET(a)      ((a)->u.d.u.thedata.offset)
#define TYPE_DECL_TYPEDATA(a)    (&((a)->u.d.u.thedata))
#define TYPE_DECL_PARAMS(a)      ((a)->u.d.u.function.param_list)
#define TYPE_DECL_FBODY(a)       ((a)->u.d.u.function.body)
#define TYPE_SPEC_TYPE(a)        ((a)->u.s.thetype)
#define TYPE_SPEC_STRUCTDEF(a)   ((a)->u.s.struct_def)
#define TYPE_STRUCTDEF_FIELDS(a) ((a)->u.s.struct_def->fields)
#define TYPE_STRUCTDEF_NAME(a)   ((a)->u.s.struct_def->name)
#define TYPE_STRUCTDEF_LEVEL(a)  ((a)->u.s.struct_def->level)
#define TYPE_IS_STRUCT(a)        (TYPE_CLASS(a) == SPECIFIER && \
                                  TYPE_SPEC_TYPE(a) & STRUCT_TYPE)
#define TYPE_IS_UNION(a)           (TYPE_CLASS(a) == SPECIFIER && \
				    TYPE_SPEC_TYPE(a) & UNION_TYPE)
#define TYPE_IS_STRUCT_OR_UNION(a) ((TYPE_CLASS(a) == SPECIFIER && \
				     TYPE_SPEC_TYPE(a) & (UNION_TYPE | \
							  STRUCT_TYPE)))
#define TYPE_IS_BASIC(a)         (TYPE_CLASS(a) == SPECIFIER && \
				  !TYPE_MODIFIERS(a) &&         \
				  !TYPE_IS_STRUCT_OR_UNION(a))
#define TYPE_IS_SIGNED(a)        ((TYPE_SPEC_TYPE(a) & SIGNED_TYPE))
#define TYPE_MODIFIERS(a)        ((a)->type_modifiers)
#define TYPE_IS_MODIFIER(a)      (TYPE_CLASS(a) == DECLARATOR && \
				  TYPE_DECL_TYPE(a) == MODIFIER_TYPE)
#define TYPE_IS_ARRAY(a)         (TYPE_CLASS(a) == DECLARATOR &&\
				  TYPE_DECL_TYPE(a) == ARRAY_TYPE)
#define TYPE_IS_POINTER(a)       (TYPE_CLASS(a) == DECLARATOR &&\
				  TYPE_DECL_TYPE(a) == POINTER_TYPE)
#define TYPE_IS_FUNCTION(a)      (TYPE_CLASS(a) == DECLARATOR &&\
				  TYPE_DECL_TYPE(a) == FUNCTION_TYPE)
#define TYPE_IS_BITFIELD(a)      (TYPE_CLASS(a) == DECLARATOR && \
                                  TYPE_DECL_TYPE(a) == BITFIELD_TYPE)
#define TYPE_IS_SPECIFIER(a)     (TYPE_CLASS(a) == SPECIFIER)
#define TYPE_IS_DECLARATOR(a)    (TYPE_CLASS(a) == DECLARATOR)
#define TYPE_IS_SIMPLE(a)        (TYPE_IS_SPECIFIER(a) && \
                                  !TYPE_IS_STRUCT_OR_UNION(a))
#define TYPE_SIZE(a)             (TYPEDATA_SIZE (TYPE_MODIFIERS (a)))
#define TYPE_OFFSET(a)           (TYPEDATA_OFFSET (TYPE_MODIFIERS (a)))    
#define TYPE_ALIGN(a)            (TYPEDATA_ALIGN (TYPE_MODIFIERS (a)))
#define TYPE_BORDER(a)           (TYPEDATA_BORDER (TYPE_MODIFIERS (a)))
#define TYPE_MEMSIZE(a)          (TYPEDATA_MEMSIZE (TYPE_MODIFIERS (a)))
#define TYPE_COMPLETE(a)         (TYPEDATA_COMPLETE (TYPE_MODIFIERS(a)))


typedef struct symbol {
  char *name;
  char *local_name;
  Type *type, *etype;
  TypeData *type_data;
  struct symbol *next, *end;
  int    level;
} Symbol;


#define SYM_NAME(a)       ((a)->name)
#define SYM_LOCAL_NAME(a) ((a)->local_name)
#define SYM_TYPE(a)       ((a)->type)
#define SYM_ETYPE(a)      ((a)->etype)
#define SYM_NEXT(a)       ((a)->next)
#define SYM_END(a)        ((a)->end)
#define SYM_LEVEL(a)      ((a)->level)
#define SYM_TYPEDATA(a)   ((a)->type_data)
#define SYM_SIZE(a)       ((a)->type_data->size)
#define SYM_MEMSIZE(a)    ((a)->type_data->memsize)
#define SYM_ALIGN(a)      ((a)->type_data->alignment)
#define SYM_BORDER(a)     ((a)->type_data->byte_order)
#define SYM_OFFSET(a)     ((a)->type_data->offset)
#define SYM_COMPLETE(a)   ((a)->type_data->complete_type)

#define SYM_IS_BITFIELD(a)         (TYPE_IS_BITFIELD(SYM_TYPE(a)))
#define SYM_IS_STRUCT_OR_UNION(a)  (TYPE_IS_STRUCT_OR_UNION(SYM_TYPE(a)))
#define SYM_IS_STRUCT(a)           (TYPE_IS_STRUCT(SYM_TYPE(a)))
#define SYM_IS_UNION(a)            (TYPE_IS_UNION(SYM_TYPE(a)))
#define SYM_IS_ARRAY(a)            (TYPE_IS_ARRAY(SYM_TYPE(a)))
#define SYM_IS_POINTER(a)          (TYPE_IS_POINTER(SYM_TYPE(a)))
#define SYM_IS_FUNCTION(a)         (TYPE_IS_FUNCTION(SYM_TYPE(a)))
#define SYM_FUNC_RETURN(a)         (TYPE_NEXT(SYM_TYPE(a)))
#define SYM_IS_SIGNED(a)           (TYPE_IS_SIGNED(SYM_TYPE(a)))
#define SYM_IS_SIMPLE_TYPE(a)      (TYPE_IS_SIMPLE(SYM_TYPE(a)))
#define SYM_NATIVE(a)              (SYM_TYPE(a) && TYPE_NATIVE (SYM_TYPE(a)))
#define SYM_ADDNATIVE(a)           (SYM_TYPE(a) && TYPE_ADDNATIVE(SYM_TYPE(a)))

/* The following type is used by sizeof to get info on variable length arrays */

typedef struct {
  int   elem_size;
  char* length_identifier;
} VarLen;

#define VAR_ARRAY_ELEM_SIZE(a)     ((a)->elem_size)
#define VAR_ARRAY_LENGTH_ID(a)     ((a)->length_identifier)

/* these aren't what the default values are, they are just what the field
   gets set to to tell it to be the default value.
*/

#define DEFAULT_SIZE             0
#define DEFAULT_ALIGN            0
#define DEFAULT_BORDER        NULL
#define DEFAULT_OFFSET     INT_MAX

#define equal_byteorder(a,b) (!strcmp((a), (b)))
#define byteorder_len(a)     (strlen(a))
#endif /* in symbol_defs.h */
